"""
# ============================================================================
# Creates a tree of all the courses, and lists of all the resources
# - can add/modify/delete all objects through the tree object
# - can drag resources onto courses
#
# Events triggered by EditCoursesTk
#       button_event_create_new_course()
#       tree_event_edit_tree_obj(selected_obj, parent_obj, tree_id)
#       tree_event_create_tree_popup(selected_obj, parent_obj, tree_id, parent_id)->list[MenuItem]
#       resource_event_create_resource_popup(resource_type, object)->list[MenuItem]
#       resource_event_show_teacher_stat(teacher)
#       resource_drag_event_is_valid_drop(resource_type, tree_object) -> bool
#       resource_dropped_event(resource_obj, tree_obj, tree_id)
#
# Events triggered by CreateTreePopupMenuActions
#       edit_block_dialog(selected_object, tree_id)
#       add_blocks_dialog(selected_object, tree_id)
#       edit_section_dialog(selected_object, tree_id)
#       edit_course_dialog(selected_object, tree_id)
#       remove_selected_from_parent(parent_object, selected_object, tree_parent_id))
#       modify_course_needs_allocation(selected_object, true_false, tree_id)
#       add_section_dialog(selected_object, tree_id)
#       assign_selected_to_parent(selected_object, resource, tree_id)
#       remove_selected_from_parent(selected_object, resource, tree_id)
#       remove_all_types_from_selected_object(resource_type_string, selected_object, tree_id)
#
# ============================================================================
"""

from functools import partial
from typing import Optional, Callable, Any, TYPE_CHECKING, Literal

from ..gui_generics.menu_and_toolbars import MenuItem
from ..gui_dialogs.add_edit_block_dialog_tk import AddEditBlockDialogTk
from ..gui_dialogs.add_section_dialog_tk import AddSectionDialogTk
from ..gui_dialogs.edit_course_dialog_tk import EditCourseDialogTk
from ..gui_dialogs.edit_section_dialog_tk import EditSectionDialogTk
from ..model import Schedule, ResourceType, Section, Block, Teacher, Lab, Stream, Course, WeekDay
from ..gui_pages import EditCoursesTk
from ..presenter.edit_courses_popup_menus import CreateTreePopupMenuActions, CreateResourcePopupMenuActions

RESOURCE_OBJECT = Teacher | Lab | Stream
TREE_OBJECT = Any

# =====================================================================================================================
# model subroutine lookups
# =====================================================================================================================
REMOVE_SUBS = {
    "teacher": lambda parent, teacher: parent.remove_teacher(teacher),
    "lab": lambda parent, lab: parent.remove_lab(lab),
    "stream": lambda parent, stream: parent.remove_stream(stream),
    "course": lambda parent, course: parent.remove_course(course),
    "block": lambda parent, block: parent.remove_block(block),
    "section": lambda parent, section: parent.remove_section(section),
}
REMOVE_ALL_SUBS = {
    "teacher": lambda parent: parent.remove_all_teachers(),
    "lab": lambda parent: parent.remove_all_labs(),
    "stream": lambda parent: parent.remove_all_streams(),
    "block": lambda parent: parent.remove_all_blocks(),
    "section": lambda parent: parent.remove_all_sections(),
}
REFRESH_SUBS = {
    "schedule": lambda presenter, parent_id, obj1: presenter.refresh(),
    "course": lambda presenter, parent_id, obj1:  presenter.refresh_course_gui(parent_id, obj1, False),
    "block": lambda presenter, parent_id, obj1: presenter.refresh_block_gui(parent_id, obj1, False),
    "section": lambda presenter, parent_id, obj1: presenter.refresh_section_gui(parent_id, obj1, False),
}
ASSIGN_SUBS = {
    "teacher": lambda parent, teacher: parent.add_teacher(teacher),
    "lab": lambda parent, lab: parent.add_lab(lab),
    "stream": lambda parent, stream: parent.add_stream(stream),
}

def list_minus_list(all_list, other_list):
    other_list = set(other_list)
    minus_list = list(set(all_list).difference(other_list))
    minus_list.sort()
    return  minus_list, list(other_list)


# =====================================================================================================================
# EditCourse
# =====================================================================================================================
class EditCourses:
    """
    Creates the page for editing courses, with a tree view, lists of resources, drag'n'drop, etc.
    """

    def __init__(self,
                 dirty_flag_method: Callable[[Optional[bool]], bool],
                 frame,
                 schedule: Optional[Schedule],
                 gui: EditCoursesTk=None):
        """
        :param dirty_flag_method: a function that is used to set the flag if schedule has been changed
        :param frame: the frame to put all the gui stuff in
        :param schedule: the model
        :param gui: the gui page
        """

        if not gui:
            self.gui = EditCoursesTk(frame)
        else:
            self.gui = gui

        self.set_dirty_flag = dirty_flag_method
        self.frame = frame
        self.schedule = schedule
        self.tree_ids: dict[str, str] = {}


        # set all the event required handlers for EditResourcesTk
        self.gui.handler_tree_edit = self.tree_event_edit_tree_obj
        self.gui.handler_new_course = self.button_event_create_new_course
        self.gui.handler_tree_create_popup = self.tree_event_create_tree_popup
        self.gui.handler_resource_create_menu = self.resource_event_create_resource_popup
        self.gui.handler_show_teacher_stat = self.resource_event_show_teacher_stat
        self.gui.handler_drag_resource = self.resource_drag_event_is_valid_drop
        self.gui.handler_drop_resource = self.resource_dropped_event

    # -----------------------------------------------------------------------------------------------------------------
    # tree - refresh everything
    # -----------------------------------------------------------------------------------------------------------------
    def refresh(self):
        """ updates the gui with the data from schedule"""
        self.gui.clear_tree()
        self.tree_ids.clear()
        for course in self.schedule.courses():
            name = " ".join((course.number, course.name))
            course_id = self.gui.add_tree_item("", name, course)
            self.refresh_course_gui(course_id, course, True)
        self.gui.update_resource_type_objects(ResourceType.teacher, self.schedule.teachers())
        self.gui.update_resource_type_objects(ResourceType.lab, self.schedule.labs())
        self.gui.update_resource_type_objects(ResourceType.stream, self.schedule.streams())


    # -----------------------------------------------------------------------------------------------------------------
    # tree - refresh course
    # -----------------------------------------------------------------------------------------------------------------
    def refresh_course_gui(self, course_id, course: Course, hide:bool = True):
        """
        refresh the contents of the course (sections) on the tree structure
        :param course_id: the id of the Course tree item
        :param course:
        :param hide: leave the parent (and everything under it hidden?)
        :return:
        """
        self.gui.remove_tree_item_children(course_id)
        for section in course.sections():
            name = str(section)
            section_id = self.gui.add_tree_item(course_id, name, section, hide)
            self.refresh_section_gui(section_id, section, hide)

    # -----------------------------------------------------------------------------------------------------------------
    # tree - refresh gui
    # -----------------------------------------------------------------------------------------------------------------
    def refresh_section_gui(self, section_id, section: Section, hide: bool = True):
        """
        adds the contents of the section (blocks) on the tree structure
        :param section_id: the id of the Section tree item
        :param section: course section
        :param hide: leave the parent (and everything under it hidden?)
        :return:
        """
        self.gui.remove_tree_item_children(section_id)

        # change the name of the parent text if stream exists for this section
        name = str(section)
        if len(section.streams()):
            name += "  (" + ",".join([str(stream) for stream in section.streams()]) + ")"
            self.gui.update_tree_text(section_id, name)

        for block in section.blocks():
            name = "Class Time: " + block.description()
            block_id = self.gui.add_tree_item(section_id, name, block, hide)
            self.refresh_block_gui(block_id, block, hide)


    # -----------------------------------------------------------------------------------------------------------------
    # tree - refresh block
    # -----------------------------------------------------------------------------------------------------------------
    def refresh_block_gui(self, block_id, block: Block, hide: bool=True):
        """
        adds the contents of the block (resources) onto the tree structure
        :param block_id: the id of the Block tree item
        :param block: course section block
        :param hide: leave the parent (and everything under it hidden?)
        :return:
        """
        self.gui.remove_tree_item_children(block_id)
        for teacher in block.teachers():
            self.gui.add_tree_item(block_id, f"teacher: {str(teacher)}", teacher, hide)
        for lab in block.labs():
            self.gui.add_tree_item(block_id, f"lab: {str(lab)}", lab, hide)

    # -------------------------------------------------------------------------------------------------------------
    # Tree - edit object
    # -------------------------------------------------------------------------------------------------------------
    def tree_event_edit_tree_obj(self, obj: Any, parent_obj: Any, tree_id: str, parent_id: str):
        """
        Given a particular tree object, edit it (Blocks, Section, Course)
        :param obj: Any tree object, but will ignore it if it is not a Block, Section or Course
        :param parent_obj: Who is the parent of this object
        :param tree_id: What is the tree_id of this object
        :param parent_id: What is the tree_id of this object's parent object
        :return:
        """
        if isinstance(obj, Block):
            self.edit_block_dialog(obj,parent_obj, parent_id)
        if isinstance(obj, Section):
            self.edit_section_dialog(obj, tree_id)
        if isinstance(obj, Course):
            self.edit_course_dialog(obj, tree_id)

    # -------------------------------------------------------------------------------------------------------------
    # Tree - popup menu
    # -------------------------------------------------------------------------------------------------------------
    def tree_event_create_tree_popup(self, selected_obj: Any, parent_object, tree_path:str, tree_parent_path) -> list[MenuItem]:
        """
        Create a pop-up menu based on the selected object

        :param selected_obj: object that was selected on the tree
        :param parent_object: parent of selected object
        :param tree_path: the id of the tree item that was selected
        :param tree_parent_path: the id of the tree item that is the parent of the selected
        :return:
        """
        popup = CreateTreePopupMenuActions(self, selected_obj, parent_object, tree_path, tree_parent_path)
        return popup.create_tree_popup_menus()

    # -------------------------------------------------------------------------------------------------------------
    # Course - add dialog
    # -------------------------------------------------------------------------------------------------------------
    def button_event_create_new_course(self):
        """
        Create a new course
        """
        self._add_edit_course_dialog('add', None, None)

    # -------------------------------------------------------------------------------------------------------------
    # Course - edit dialog
    # -------------------------------------------------------------------------------------------------------------
    def edit_course_dialog(self, course: Course, tree_id:str):
        """
        Create and use the edit course dialog to modify the selected course
        :param course:
        :param tree_id:
        """
        self._add_edit_course_dialog('edit',course,tree_id)

    # -------------------------------------------------------------------------------------------------------------
    # Course - add or edit dialog
    # -------------------------------------------------------------------------------------------------------------
    def _add_edit_course_dialog(self, add_or_edit: Literal['add','edit'], course: Optional[Course], tree_id:Optional[str]):
        """
        Create and use the edit course dialog to modify the selected course
        :param course:
        :param tree_id:
        """

        # set defaults
        block_data = []

        non_assigned_teachers, assigned_teachers = list_minus_list(self.schedule.teachers(), [])
        non_assigned_labs, assigned_labs = list_minus_list( self.schedule.labs(), [])

        course_number = ""
        number_sections = 1
        hours_per_week = 3
        course_name = ""
        course_allocation = True

        # if course exists, get info from course
        if course is not None:
            course_number = course.number
            number_sections = len(course.sections())
            hours_per_week = course.hours_per_week
            course_name = course.name
            course_allocation = course.needs_allocation

            if len(course.sections()) != 0:
                for b in course.sections()[0].blocks():
                    block_data.append((b.day.name, b.start, b.duration))

            non_assigned_teachers, assigned_teachers = list_minus_list(self.schedule.teachers(), course.teachers())
            non_assigned_labs, assigned_labs = list_minus_list(self.schedule.labs(), course.labs())

        # open the dialog box
        EditCourseDialogTk(self.frame,
                                course_number=course_number,
                                edit_or_add=add_or_edit,
                                existing_course_numbers=[c.number for c in self.schedule.courses()],
                                course_name=course_name,
                                course_hours=hours_per_week,
                                course_allocation= course_allocation,
                                num_sections=number_sections,

                                assigned_teachers=assigned_teachers,
                                non_assigned_teachers=non_assigned_teachers,
                                assigned_labs=assigned_labs,
                                non_assigned_labs=non_assigned_labs,
                                current_blocks=block_data,
                                apply_changes=self.schedule.add_edit_course,
                                )

        # refresh
        if add_or_edit == 'edit':
            REFRESH_SUBS['course'](self, tree_id, course)
        else:
            REFRESH_SUBS['schedule'](self, None, None)
        self.set_dirty_flag(True)


    # -------------------------------------------------------------------------------------------------------------
    # Section - add dialog
    # -------------------------------------------------------------------------------------------------------------
    def add_section_dialog(self, course: Course, tree_id: str):
        """Add section to Course
        :param course:
        :param tree_id:
        """
        title = f"{course.name} ({course.hours_per_week} hrs)"
        AddSectionDialogTk(self.frame,
                                course_description=title,
                                apply_changes=partial(self.schedule.add_sections,course),
                                course_hours=course.hours_per_week),
        REFRESH_SUBS["course"](self, tree_id, course)
        self.set_dirty_flag(True)

    # -------------------------------------------------------------------------------------------------------------
    # Section - edit dialog
    # -------------------------------------------------------------------------------------------------------------
    def edit_section_dialog(self, section: Section, tree_id: str):
        """Modify an existing section
        :param section:
        :param tree_id:
        """

        title = f"{section.course.name} ({section.course.hours_per_week} hrs)"
        text = section.title
        block_data = []
        for b in section.blocks():
            block_data.append((b.day.name, b.start, b.duration))
        non_assigned_teachers, assigned_teachers = list_minus_list(self.schedule.teachers(),section.teachers())
        non_assigned_labs, assigned_labs = list_minus_list(self.schedule.labs(),section.labs())
        non_assigned_streams, assigned_streams = list_minus_list(self.schedule.streams(), section.streams())

        EditSectionDialogTk(self.frame,
                                 course_description=title,
                                 section_description=text,
                                 assigned_teachers=list(section.teachers()),
                                 non_assigned_teachers=non_assigned_teachers,
                                 assigned_labs=list(section.labs()),
                                 non_assigned_labs=non_assigned_labs,
                                 assigned_streams=list(section.streams()),
                                 non_assigned_streams=non_assigned_streams,
                                 current_blocks=block_data,
                                 apply_changes=partial(self.schedule.update_section, section),
                                 course_hours=section.course.hours_per_week)
        REFRESH_SUBS["section"](self, tree_id, section)

        self.set_dirty_flag(True)

    # -------------------------------------------------------------------------------------------------------------
    # Blocks - add dialog
    # -------------------------------------------------------------------------------------------------------------
    def add_blocks_dialog(self, section: Section, tree_id: str):
        """
        :param section: the section to add a block to
        :param tree_id: the id of the section
        """
        AddEditBlockDialogTk(self.frame, "add", 1.5, [], list(self.schedule.teachers()),
                             [], list(self.schedule.labs()), partial(self.schedule.add_blocks, section))

        REFRESH_SUBS["section"](self, tree_id, section)
        self.set_dirty_flag(True)

    # -------------------------------------------------------------------------------------------------------------
    # Blocks - edit dialog
    # -------------------------------------------------------------------------------------------------------------
    def edit_block_dialog(self, block, section: Any, parent_id: str):
        """
        Modify the block
        :param block:
        :param section:
        :param parent_id: the id of the section
        """
        non_assigned_teachers, assigned_teachers = list_minus_list(self.schedule.teachers(), block.teachers())
        non_assigned_labs, assigned_labs = list_minus_list(self.schedule.labs(), block.labs())

        AddEditBlockDialogTk(
            frame = self.frame,
            add_edit_type = "edit",
            duration = block.duration,
            assigned_teachers = list(block.teachers()),
            non_assigned_teachers= non_assigned_teachers,
            non_assigned_labs=non_assigned_labs,
            assigned_labs=assigned_labs,
            apply_changes= partial(self.schedule.edit_block, block),
        )

        REFRESH_SUBS["section"](self,parent_id, block.section)
        self.set_dirty_flag(True)


    # -------------------------------------------------------------------------------------------------------------
    # Tree Popup - remove selected element from its parent
    # -------------------------------------------------------------------------------------------------------------
    def remove_selected_from_parent(self, parent, selected, parent_id):
        """
        Tree Popup - remove selected element from its parent
        :param parent:  the parent object
        :param selected: the selected object
        :param parent_id: the tree id of the parent
        """
        obj_type = str(type(selected)).lower()
        key = obj_type.split(".")[-1][0:-2]
        REMOVE_SUBS[key](parent,selected)

        obj_type = str(type(parent)).lower()
        key = obj_type.split(".")[-1][0:-2]
        REFRESH_SUBS[key](self,parent_id, parent)
        self.set_dirty_flag(True)

    # -------------------------------------------------------------------------------------------------------------
    # Tree Popup - needs allocation
    # -------------------------------------------------------------------------------------------------------------
    def modify_course_needs_allocation(self, course:Course, value: bool, tree_path):
        """
        set the 'needs allocation' property of the course
        :param course:
        :param value:
        :param tree_path:
        """
        course.needs_allocation = value
        self.refresh_course_gui(tree_path, course)
        self.set_dirty_flag(True)

    # -------------------------------------------------------------------------------------------------------------
    # Tree Popup or dropped object - assign resource to selected tree item
    # -------------------------------------------------------------------------------------------------------------
    def assign_selected_to_parent(self, parent, selected, parent_id = -1):
        """
        assign selected resource object to the parent
        :param parent: Course/Section/block object
        :param selected: Lab/Stream/Teacher object
        :param parent_id: tree id of the parent object
        :return:
        """
        obj_type = str(type(selected)).lower()
        key = obj_type.split(".")[-1][0:-2]
        ASSIGN_SUBS[key](parent,selected)

        if parent_id == -1:
            key = 'schedule'
        else:
            obj_type = str(type(parent)).lower()
            key = obj_type.split(".")[-1][0:-2]
        REFRESH_SUBS[key](self,parent_id, parent)
        self.set_dirty_flag(True)


    # -------------------------------------------------------------------------------------------------------------
    # remove all resources from a selected object
    # -------------------------------------------------------------------------------------------------------------
    def remove_all_types_from_selected_object(self, obj_type, selected, selected_id):
        """
        remove all resources from a selected object
        :param obj_type: the type of the object (Course/Section/Block)
        :param selected: the object
        :param selected_id: the tree id of the selected object
        :return:
        """
        REMOVE_ALL_SUBS[obj_type](selected)
        obj_type = str(type(selected)).lower()
        key = obj_type.split(".")[-1][0:-2]
        REFRESH_SUBS[key](self, selected_id, selected)
        self.set_dirty_flag(True)

    # -------------------------------------------------------------------------------------------------------------
    # create a pop-up menu for the resource list item
    # -------------------------------------------------------------------------------------------------------------
    def resource_event_create_resource_popup(self, resource_type: ResourceType, resource: RESOURCE_OBJECT) -> list[MenuItem]:
        """
        Create a pop-up menu based on the selected resource object

        :param resource_type: is it a teacher/lab/stream?
        :param resource: selected object
        """
        popup = CreateResourcePopupMenuActions(self, self.schedule, resource_type, resource)
        return popup.create_resource_popup_menus()


    # -------------------------------------------------------------------------------------------------------------
    # respond to a double click on a teacher resource
    # -------------------------------------------------------------------------------------------------------------
    def resource_event_show_teacher_stat(self, teacher: Teacher):
        text = []
        days = set()
        hours = 0
        for course in self.schedule.get_courses_for_teacher(teacher):
            for block in course.blocks():
                if block.has_teacher(teacher):
                    days.add(block.day)
                    hours = hours + block.duration

        text.append(f"Release: {teacher.release}")
        text.append( f"Total Class Time: {hours}")
        day_str = []
        for day in WeekDay:
            if day in days:
                day_str.append(day.name[0:3])
        text.append( f"Days: "+", ".join(day_str))
        self.gui.show_message(title = "Teacher Stats", msg=str(teacher), detail="\n".join(text))

    # -------------------------------------------------------------------------------------------------------------
    # can the resource object be added to the selected tree object?
    # -------------------------------------------------------------------------------------------------------------
    def resource_drag_event_is_valid_drop(self, resource_type: ResourceType, destination: TREE_OBJECT) -> bool:
        """
        is the destination object a legitimate object to accept as a parent for this resource type?
        :param resource_type:
        :param destination:
        """
        match resource_type:
            case ResourceType.teacher:
                if isinstance(destination, Course) or isinstance(destination, Section) or isinstance(destination, Block):
                    return True
                else:
                    return False
            case ResourceType.stream:
                if isinstance(destination, Section):
                    return True
                else:
                    return False
            case ResourceType.lab:
                if isinstance(destination, Course) or isinstance(destination, Section) or isinstance(destination, Block):
                    return True
                else:
                    return False

        return False

    # -------------------------------------------------------------------------------------------------------------
    # a resource object has been dropped on a course object
    # -------------------------------------------------------------------------------------------------------------
    def resource_dropped_event(self, resource: RESOURCE_OBJECT, destination: TREE_OBJECT, tree_id):
        """
        a resource object has been dropped onto a course object, so assign this resource to this object
        :param resource:
        :param destination:
        :param tree_id:
        :return:
        """
        self.assign_selected_to_parent(destination, resource,tree_id)



"""Sets up the standard Tk properties (fonts, colors)"""

from __future__ import annotations

import os
from tkinter import Tk, TclError, ttk
from tkinter.font import Font
import platform
import re
from typing import Literal, Optional

import tkinter.font as tkFont

from PIL import Image, ImageDraw, ImageTk

from ..Utilities.Colour import get_colour_string_from_rgb, darken, is_light, lighten, add, \
    hsl, get_colour_string_from_hsl

operating_system = platform.system().lower()

DEFAULT_FONT_SIZE = 10
system="win"
if "darwin" in operating_system:
    DEFAULT_FONT_SIZE = 13
    system="mac"

global img_open, img_close, img_empty


class TkColours:
    """Defines colour scheme """

    def __init__(self, tk_root: Optional[Tk] = None, dark_mode=False):

        window_background_color: Optional[str] = None
        pressed_button_text_color: Optional[str] = None
        text_color: Optional[str] = None
        text_background_color: Optional[str] = None
        selected_text_background_color: Optional[str] = None
        selected_text_color: Optional[str] = None

        # ============================================================================
        # get system colours (mac) ... not sure about if this works on windows
        # ============================================================================
        mac = "mac"
        win = "win"
        default_colour_names = {
            "window_background_color": {mac:'systemWindowBackgroundColor', win:'SystemButtonFace'},
            "pressed_button_text_color":{mac:'systemPressedButtonTextColor', win:'SystemButtonFace'},
            "text_color":{mac:'systemTextColor', win:'SystemButtonText'},
            "text_background_color":{mac:'systemTextBackgroundColor', win:'SystemButtonFace'},
            "selected_text_background_color":{mac:'systemSelectedTextBackgroundColor', win:'SystemHighlight'},
            "selected_text_color":{mac:'systemSelectedTextColor', win:'SystemHighlightText'},
        }
        if tk_root:
            try:
                (r, g, b) = tk_root.winfo_rgb(default_colour_names['window_background_color'][system])
                window_background_color = get_colour_string_from_rgb(r / 65536.0, g / 65536.0, b / 65536.0)
            except TclError:
                pass

            try:
                (r, g, b) = tk_root.winfo_rgb(default_colour_names['pressed_button_text_color'][system])
                pressed_button_text_color = get_colour_string_from_rgb(r / 65536.0, g / 65536.0, b / 65536.0)
            except TclError:
                pass

            try:
                (r, g, b) = tk_root.winfo_rgb(default_colour_names['text_color'][system])
                text_color = get_colour_string_from_rgb(r / 65536.0, g / 65536.0, b / 65536.0)
            except TclError:
                pass

            try:
                (r, g, b) = tk_root.winfo_rgb(default_colour_names['text_background_color'][system])
                text_background_color = get_colour_string_from_rgb(r / 65536.0, g / 65536.0, b / 65536.0)
            except TclError:
                pass

            try:
                (r, g, b) = tk_root.winfo_rgb(default_colour_names['selected_text_background_color'][system])
                selected_text_background_color = get_colour_string_from_rgb(r / 65536.0, g / 65536.0,
                                                                                   b / 65536.0)
            except TclError:
                pass

            try:
                (r, g, b) = tk_root.winfo_rgb(default_colour_names['selected_text_color'][system])
                selected_text_color = get_colour_string_from_rgb(r / 65536.0, g / 65536.0, b / 65536.0)
            except TclError:
                pass

        # default is mac
        self.SelectedForeground: str = selected_text_color if selected_text_color else "#000000"
        self.SelectedBackground: str = selected_text_background_color if selected_text_background_color \
            else "#eab9ff"
        self.DataBackground = text_background_color if text_background_color else "#ffffff"
        self.WorkspaceColour: str = window_background_color if window_background_color else "#eeeeee"
        self.ButtonPress: str = pressed_button_text_color if pressed_button_text_color else "#ffffff"
        self.DataForeground = text_color if text_color else "#000000"
        self.HighlightColor = self.DataForeground
        self.WindowForeground: str = self.DataForeground
        self.ButtonBackground: str = self.WorkspaceColour
        self.ActiveBackground: str = self.WorkspaceColour
        self.HighlightBackground: str = self.WorkspaceColour
        self.DarkBackground: str = darken(self.WorkspaceColour, 20)
        self.DisabledBackground: str = "#a3a3a3"
        self.DirtyColour: str = "#ff8888"
        self.DisabledBackground = "#a3a3a3"
        self.DisabledForeground = "#000000"
        self.DragNDropForeground = self.WindowForeground

        if is_light(self.WorkspaceColour):
            self.DragNDropBackground = darken(self.WorkspaceColour, 25)
            self.DragNDropAllowedBackground = darken(self.WorkspaceColour, 5)
        else:
            self.DragNDropBackground = lighten(self.WorkspaceColour, 25)
            self.DragNDropAllowedBackground = lighten(self.WorkspaceColour, 5)

        self.ButtonHighlight = self.ButtonBackground
        if is_light(self.ButtonHighlight):
            self.ButtonHoverHighlight = darken(self.WorkspaceColour, 25)
            self.ButtonForeground = "#000000"
        else:
            self.ButtonHoverHighlight = lighten(self.WorkspaceColour, 25)
            self.ButtonForeground = "#FFFFFF"

        if is_light(self.WorkspaceColour):
            self.DirtyColour = add("red","#444444")
        else:
            self.DirtyColour = add("red","white")

        # ============================================================================
        # if invert mode, just invert all the colours
        # ============================================================================
        if ((dark_mode and is_light(self.WorkspaceColour)) or
                (not dark_mode and not is_light(self.WorkspaceColour))):

            for col in vars(self):
                # Note: button colors don't work on MAC, so skip
                if re.search("button", col.lower()):
                    if re.match('darwin', operating_system):
                        continue

                h, s, l = hsl(getattr(self, col))
                l = abs(0.9 * l - 1)
                setattr(self, col, get_colour_string_from_hsl(h, s, l))

        if os.name == 'darwin' or os.name == "posix":
            # MAC does not allow you to change the background colour of buttons
            self.ButtonForeground = "#000000"

    def __str__(self):
        str: str = ""
        for col in vars(self):
            str += f"{col}: {getattr(self, col)}\n"
        return str


class TkFonts:
    def __init__(self, mw: Tk, my_size: int = DEFAULT_FONT_SIZE):
        default_font = tkFont.nametofont("TkTextFont")
        family = default_font["family"]

        normal_weight: Literal['normal'] = 'normal'
        bold_weight: Literal['bold'] = 'bold'
        normal_size = my_size
        bigger_size = my_size + 2
        normal_font = family
        fixed_font = 'courier new'

        # make fonts
        self.normal: Font = Font(mw, weight=normal_weight, size=normal_size, family=normal_font)
        self.bold: Font = Font(mw, weight=bold_weight, size=normal_size, family=normal_font)
        self.big: Font = Font(mw, weight=normal_weight, size=bigger_size, family=normal_font)
        self.bigbold: Font = Font(mw, weight=bold_weight, size=bigger_size, family=normal_font)
        self.fixed: Font = Font(mw, weight=bold_weight, size=my_size + 1, family=fixed_font)
        self.small: Font = Font(mw, weight=normal_weight, size=my_size - 2, family=normal_font)


colours: TkColours = TkColours()
fonts: Optional[TkFonts] = None


def get_fonts_and_colours():
    return colours, fonts


def set_default_fonts(mw: Tk, font_size: Optional[int] = DEFAULT_FONT_SIZE):
    global fonts
    mw.configure(background=colours.WorkspaceColour)

    if font_size is None:
        font_size = DEFAULT_FONT_SIZE
    fonts = TkFonts(mw, font_size)
    mw.option_add("*Font", fonts.normal, )
    return fonts

def set_default_fonts_and_colours(mw: Tk, font_size: Optional[int] = DEFAULT_FONT_SIZE, invert: bool = False):
    global fonts
    global colours
    colours = TkColours(mw, dark_mode=invert)
    mw.configure(background=colours.WorkspaceColour)

    if font_size is None:
        font_size = DEFAULT_FONT_SIZE
    fonts = TkFonts(mw, font_size)
    mw.option_add("*Font", fonts.normal, )
    return colours, fonts


# =================================================================
# set_system_colours
# =================================================================
def set_system_colours(mw: Tk, colors: TkColours, theme: str):
    """Using Colour array and main window, set_default_fonts_and_colours up
    some standard defaults for Tk widgets"""

    # generic
    #    activebackground − Background color for the widget when the widget is active.
    #    activeforeground − Foreground color for the widget when the widget is active.
    #    background − Background color for the widget. This can also be represented as bg.
    #    disabledforeground − Foreground color for the widget when the widget is disabled.
    #    highlightbackground − Background color of the highlight region when the widget has focus.
    #    highlightcolor − Foreground color of the highlight region when the widget has focus.
    #    foreground − Foreground color for the widget. This can also be represented as fg.
    #    selectbackground − Background color for the selected items of the widget.
    #    selectforeground − Foreground color for the selected items of the widget.
    _option_add(mw, '*activebackground', colors.ActiveBackground)
    _option_add(mw, '*activeforeground', colors.ButtonPress)
    _option_add(mw, "*background", colors.WorkspaceColour)
    _option_add(mw, "*disabledforeground", colors.DisabledForeground)
    _option_add(mw, "*highlightbackground", colors.HighlightBackground)
    _option_add(mw, "*highlightcolor", colors.HighlightColor)
    _option_add(mw, '*foreground', colors.WindowForeground)
    _option_add(mw, '*selectBackground', colors.SelectedBackground)
    _option_add(mw, '*selectForeground', colors.SelectedForeground)
    _option_add(mw, '*troughColor', colors.DarkBackground)
    _option_add(mw, '*readonlybackground', colours.WorkspaceColour)


    # buttons
    _option_add(mw, "*Button.background", colors.ButtonBackground)
    _option_add(mw, "*Button.foreground", colors.ButtonForeground)
    _option_add(mw, "*Button.activeBackground", colors.ActiveBackground)
    _option_add(mw, "*Button.activeForeground", colors.ButtonForeground)
    _option_add(mw, "*Button.highlightBackground", colors.HighlightBackground)

    # radio buttons
    _option_add(mw, "*Radiobutton.activeBackground", colors.DarkBackground)
    _option_add(mw, "*Radiobutton.foreground", colors.WindowForeground)
    _option_add(mw, "*Radiobutton.background", colors.WorkspaceColour)
    _option_add(mw, "*Radiobutton.highlightBackground", colors.WorkspaceColour)
    _option_add(mw, "*Radiobutton.activeForeground", colors.WindowForeground)
    if re.match('windows', operating_system):
        _option_add(mw, "*Radiobutton.selectColor", colors.DataBackground)

    # check buttons
    _option_add(mw, '*Checkbutton.activeBackground', colors.DarkBackground)
    _option_add(mw, '*Checkbutton.foreground', colors.WindowForeground)
    _option_add(mw, '*Checkbutton.background', colors.WorkspaceColour)
    _option_add(mw, '*Checkbutton.activeForeground', colors.WindowForeground)
    _option_add(mw, '*Checkbutton.highlightBackground', colors.WorkspaceColour)
    if re.match('windows', operating_system):
        _option_add(mw, '*Checkbutton.selectColor', colors.DataBackground)

    # lists and entries
    _option_add(mw, '*Entry.foreground', colors.DataForeground)
    _option_add(mw, '*Entry.background', colors.DataBackground)
    _option_add(mw, '*Entry.disabledforeground', colors.DisabledForeground)
    _option_add(mw, '*Entry.disabledbackground', colors.DisabledBackground)
    _option_add(mw, '*Entry.highlightebackground', colors.WorkspaceColour)
    _option_add(mw, '*Listbox.foreground', colors.DataForeground)
    _option_add(mw, '*Listbox.background', colors.DataBackground)
    _option_add(mw, '*BrowseEntry.foreground', colors.DataForeground)
    _option_add(mw, '*BrowseEntry.background', colors.DataBackground)
    _option_add(mw, '*Entry.readonlybackground', colours.WorkspaceColour)

    # menu
    _option_add(mw, '*Menu.activeBackground', colors.DarkBackground)
    _option_add(mw, '*Menu.activeForeground', colors.WindowForeground)

    # trees
    _option_add(mw, '*DynamicTree.foreground', colors.DataForeground)
    _option_add(mw, '*DynamicTree.background', colors.DataBackground)
    _option_add(mw, '*EasyDir.foreground', colors.DataForeground)
    _option_add(mw, '*EasyDir.background', colors.DataBackground)

    # text boxes
    _option_add(mw, '*Text.foreground', colors.DataForeground)
    _option_add(mw, '*Text.background', colors.DataBackground)
    _option_add(mw, '*ROText.foreground', colors.DataForeground)
    _option_add(mw, '*ROText.background', colors.DataBackground)

    # _notebook
    _option_add(mw, "*NoteBook.inactiveBackground", colors.DarkBackground)
    _option_add(mw, "*NoteBook.background", colors.WorkspaceColour)
    _option_add(mw, "*NoteBook.backpagecolour", colors.WorkspaceColour)

    # other
    _option_add(mw, '*Scrollbar.troughColor', colors.DarkBackground)

    style = ttk.Style(mw)
    if theme is None or theme not in style.theme_names():
        if mw.tk.call('tk', 'windowingsystem') == "aqua":
            style.theme_use("aqua")
        elif mw.tk.call('tk','windowingsystem') == "win32":
            style.theme_use("winnative")
        else:
            style.theme_use("classic")
    else:
        style.theme_use(theme)

    set_treeview_style(mw, style)
    set_notebook_style(mw, style)

def set_notebook_style(root: Tk, style: ttk.Style):
    """style the notebook"""
    # https://stackoverflow.com/questions/23038356/change-color-of-tab-header-in-ttk-notebook
    if fonts is None:
        set_default_fonts_and_colours(root)
    if is_light(colours.WorkspaceColour):
        selected = darken(colours.WorkspaceColour, 20)
    else:
        selected = darken(colours.WorkspaceColour, 20)

    style.configure("TNotebook", tabmargins= [2, 5, 2, 0])
    style.configure("TNotebook.Tab", padding=[5,1], background=colours.WorkspaceColour,
                    foreground=colours.WindowForeground)
    style.map("TNotebook.Tab", background=[("selected", selected)],
              expand=[("selected", [1, 1, 1, 1])])


def set_treeview_style(root: Tk, style: ttk.Style):
    """style the Treeview"""
    # https://stackoverflow.com/questions/61280744/tkinter-how-to-adjust-treeview-indentation-size-and-indicator-arrow-image
    global img_open, img_close, img_empty


    size = 15
    if fonts is not None:
        size = 2 + fonts.normal.actual("size")


    # custom indicator images
    im_open = Image.new('RGBA', (size, size), colours.WorkspaceColour)
    im_empty = Image.new('RGBA', (size, size), colours.WorkspaceColour)
    draw = ImageDraw.Draw(im_open)
    draw.polygon([(0, 4), ((size - 1), 4), (int(size / 2), size - 4)],
                 fill=colours.DataForeground, outline=colours.DataForeground)
    im_close = im_open.rotate(90)

    img_open = ImageTk.PhotoImage(im_open, name='img_open', master=root)
    img_close = ImageTk.PhotoImage(im_close, name='img_close', master=root)
    img_empty = ImageTk.PhotoImage(im_empty, name='img_empty', master=root)

    # spacing between elements
    style.configure("Treeview", rowheight=2*size)

    # colours
    style.map("Treeview",
              background=[("selected", colours.SelectedBackground),("!selected", colours.WorkspaceColour)],
              foreground=[("selected", colours.SelectedForeground),("!selected", colours.DataForeground)],
              fieldbackground=colours.DataBackground)

    # custom indicator
    style.element_create('Treeitem.myindicator',
                         'image', 'img_close', ('user1', '!user2', 'img_open'), ('user2', 'img_empty'),
                         sticky='w', width=15)

    # replace Treeitem.indicator by custom one
    style.layout('Treeview.Item',
                 [('Treeitem.padding',
                   {'sticky': 'nswe',
                    'children': [('Treeitem.myindicator', {'side': 'left', 'sticky': ''}),
                                 ('Treeitem.image', {'side': 'left', 'sticky': ''}),
                                 ('Treeitem.focus',
                                  {'side': 'left',
                                   'sticky': '',
                                   'children': [('Treeitem.text', {'side': 'left', 'sticky': ''})]})]})]
                 )


def _option_add(mw: Tk, option: str, new_value):
    """Sets the provided option on mw. If new_value is none, does nothing"""
    if new_value:
        mw.option_add(option, new_value)



from __future__ import annotations
import configparser as cp
import os
import platform
import re
from typing import *

VALID_SEMESTER = Literal['fall', 'summer', 'winter']
DATA_FILE = "preferences.ini"
APP_DATA_PATH = None


class Preferences:
    def __init__(self):
        self._config: cp.ConfigParser = _read_ini()
        pass

    # ---------------------------------------------------------------------------------------------
    # directories
    # ---------------------------------------------------------------------------------------------
    def home_directory(self):
        """the users home directory, used to find the default place to save documents
        MAC:        $HOME
        WINDOWS:    %CSIDL_MYDOCUMENTS% or "~/Documents"
        """
        operating_system = platform.system().lower()

        user_base_dir = None
        if re.match(r'win', operating_system):
            user_base_dir = os.environ.get('CSIDL_MYDOCUMENTS', None)
            if user_base_dir is None:
                user_base_dir = os.path.expanduser("~/Documents")
        elif re.match(r'darwin|posix', operating_system):
            user_base_dir = os.environ['HOME']
        if user_base_dir is not None and os.path.isdir(user_base_dir):
            return user_base_dir
        return None

    def previous_file(self, file: Optional[str] = None) -> Optional[str]:
        """saves the filename last used for the specified semester"""
        if 'MOST_RECENT' not in self._config:
            self._config['MOST_RECENT'] = {}

        if file is not None and os.path.isfile(file):
            directory = os.path.dirname(os.path.realpath(file))
            filename = os.path.basename(file)
            if self.semester() == 'fall':
                self._config['MOST_RECENT']['fall_directory'] = directory
                self._config['MOST_RECENT']['fall_last_file'] = filename
            elif self.semester() == 'winter':
                self._config['MOST_RECENT']['winter_directory'] = directory
                self._config['MOST_RECENT']['winter_last_file'] = filename
            elif self.semester() == 'summer':
                self._config['MOST_RECENT']['summer_directory'] = directory
                self._config['MOST_RECENT']['summer_last_file'] = filename
            else:
                self._config['MOST_RECENT']['fall_directory'] = directory
                self._config['MOST_RECENT']['fall_last_file'] = filename

        else:
            if self.semester() == 'fall':
                directory = self._config['MOST_RECENT'].get('fall_directory', None)
                filename = self._config['MOST_RECENT'].get('fall_last_file', None)
            elif self.semester() == 'winter':
                directory = self._config['MOST_RECENT'].get('winter_directory', None)
                filename = self._config['MOST_RECENT'].get('winter_last_file', None)
            elif self.semester() == 'summer':
                directory = self._config['MOST_RECENT'].get('summer_directory', None)
                filename = self._config['MOST_RECENT'].get('summer_last_file', None)
            else:
                directory = self._config['MOST_RECENT'].get('fall_directory', None)
                filename = self._config['MOST_RECENT'].get('fall_last_file', None)

        if directory is not None and filename is not None:
            if os.path.isfile(directory + "/" + filename):
                y = os.path.realpath(directory + "/" + filename)
                return os.path.realpath(directory + "/" + filename)
        return None

    def current_dir(self, new_dir: Optional[str] = None) -> Optional[str]:
        """what was the last directory used to save or read a file?"""
        if 'MOST_RECENT' not in self._config:
            self._config['MOST_RECENT'] = {}

        directory = None

        if new_dir is not None and os.path.isdir(new_dir):
            directory = os.path.dirname(os.path.dirname(os.path.realpath(new_dir)))
            if self.semester() == 'fall':
                self._config['MOST_RECENT']['fall_directory'] = directory
            elif self.semester() == 'winter':
                self._config['MOST_RECENT']['winter_directory'] = directory
            elif self.semester() == 'summer':
                self._config['MOST_RECENT']['summer_directory'] = directory

        else:
            if self.semester() == 'fall':
                directory = self._config['MOST_RECENT'].get('fall_directory', None)
            if self.semester() == 'winter':
                directory = self._config['MOST_RECENT'].get('winter_directory', None)
            if self.semester() == 'summer':
                directory = self._config['MOST_RECENT'].get('summer_directory', None)

        if directory is not None:
            if os.path.isdir(directory):
                return os.path.realpath(directory)
        return None


    # ---------------------------------------------------------------------------------------------
    # look and feel
    # ---------------------------------------------------------------------------------------------
    def _set_default_theme(self):
        """set up default themes based on OS
        MAC:        'aqua', no dark mode (tk picks up on dark mode settings)
        WINDOWS:    'winnative', no dark mode (unfortunately, tk does NOT pick up on dark mode settings)
        """
        operating_system = platform.system().lower()
        if re.match(r'win', operating_system):
            self._config['COLOURS']['theme'] = "winnative"
        elif re.match(r'darwin|posix', operating_system):
            self._config['COLOURS']['theme'] = "aqua"
        else:
            self._config['COLOURS']['theme'] = "classic"

    def available_themes(self) -> tuple[str,...]:
        return 'aqua', 'clam', 'alt', 'default', 'classic', 'winnative'

    def theme(self, theme: str = None) -> str:
        """set the theme"""
        if 'COLOURS' not in self._config or 'theme' not in self._config['COLOURS']:
            self._set_default_theme()

        operating_system = platform.system().lower()

        # theme was specified
        if theme is not None:

            # ignore if not in list
            if theme in self.available_themes():

                # don't set themes that don't go with the operating system
                if re.match(r'win', operating_system) and theme != "aqua":
                    self._config['COLOURS']['theme'] = theme
                elif re.match(r'darwin|posix', operating_system) and theme != "winnative":
                    self._config['COLOURS']['theme'] = theme
                elif theme != "aqua" and theme != 'winnative':
                    self._config['COLOURS']['theme'] = theme

        return self._config['COLOURS']['theme']

    def dark_mode(self, dark_mode: bool = None) -> bool:
        """dark mode is ignored if theme is either aqua or winnative"""

        if 'COLOURS' not in self._config or 'dark_mode' not in self._config['COLOURS']:
            self._config['COLOURS'] = {'dark_mode': '0'}

        if dark_mode is not None:
            self._config['COLOURS']['dark_mode'] = str(dark_mode)
        return self._config['COLOURS'].getboolean('dark_mode')

    def font_size(self, value: Optional[int] = None) -> int:
        """Specify the _normal_ font size"""

        operating_system = platform.system().lower()
        if 'FONT_SIZE' not in self._config:
            self._config['FONT_SIZE'] = {'windows':'13', 'darwin':'10'}


        if re.match(r'win', operating_system):
            if value is not None:
                self._config['FONT_SIZE']['windows'] = str(value)
            return int(self._config['FONT_SIZE'].get("windows", '10'))

        else:
            if value is not None:
                self._config['FONT_SIZE']['darwin'] = str(value)
            return int(self._config['FONT_SIZE'].get("darwin", '13'))

    # ---------------------------------------------------------------------------------------------
    # semester
    # ---------------------------------------------------------------------------------------------
    def semester(self, semester: VALID_SEMESTER = None) -> VALID_SEMESTER:
        if 'MOST_RECENT' not in self._config:
            self._config['MOST_RECENT'] = {'semester':'fall'}
        if semester is not None:
            self._config['MOST_RECENT']['semester'] = semester
        return self._config['MOST_RECENT'].get('semester', "fall")


    # ---------------------------------------------------------------------------------------------
    # auto save
    # ---------------------------------------------------------------------------------------------
    def auto_save(self, value: Optional[bool] = None) -> bool:
        if 'AUTO_SAVE' not in self._config:
            self._config['AUTO_SAVE'] = {'set':'1'}
        if value is not None:
            if value:
                self._config['AUTO_SAVE']['set'] = '1'
            else:
                self._config['AUTO_SAVE']['set'] = '0'

        if self._config['AUTO_SAVE']['set'] == '1':
            return True
        else:
            return False


    # ---------------------------------------------------------------------------------------------
    # save the current preferences
    # ---------------------------------------------------------------------------------------------
    def save(self):
        _write_ini(self._config)


def _read_ini() -> cp.ConfigParser:
    """if the ini file can be found, read it, if the config file cannot be found, use defaults"""
    ini_file = _get_app_data_location(DATA_FILE)
    config = cp.ConfigParser()

    # ini file exists, read it
    if ini_file and os.path.exists(ini_file):
        config.read(ini_file)
        return config

    # if ini file doesn't exist, create default info
    config['MOST_RECENT'] = {}

    config['MOST_RECENT']['fall_directory'] = ''
    config['MOST_RECENT']['fall_last_file'] = ''
    config['MOST_RECENT']['winter_directory'] = ''
    config['MOST_RECENT']['winter_file'] = ''
    config['MOST_RECENT']['summer_directory'] = ''
    config['MOST_RECENT']['summer_last_file'] = ''
    config['MOST_RECENT']['semester'] = 'fall'
    config['COLOURS'] = {}
    config['COLOURS']['dark_mode'] = "yes"
    config['AUTO_SAVE'] = {}
    config['AUTO_SAVE']['set'] = '1'
    config['FONT_SIZE'] = {}
    config['FONT_SIZE']['windows'] = '13'
    config['FONT_SIZE']['darwin'] = '10'
    return config


def _get_app_data_location(ini_file_name: str) -> Optional[str]:
    global APP_DATA_PATH
    if APP_DATA_PATH is not None:
        return APP_DATA_PATH
    operating_system = platform.system().lower()
    sub_dir = "scheduling_and_allocation"

    user_base_dir = None
    app_dir = None
    if re.match(r'win', operating_system):
        user_base_dir = os.environ['APPDATA'] if os.environ['AppData'] else os.environ['USERPROFILE']
        app_dir = os.path.join(user_base_dir, sub_dir)
    elif re.search('darwin', operating_system):
        user_base_dir = os.environ['HOME']
        app_dir = os.path.join(user_base_dir, "." + sub_dir)

    if not (app_dir and os.path.exists(app_dir)):
        try:
            os.mkdir(app_dir)
        except FileNotFoundError:
            return None

    ini_file = os.path.join(user_base_dir, app_dir, ini_file_name)
    APP_DATA_PATH = ini_file
    return ini_file


def _write_ini(config: cp.ConfigParser):
    """write _preferences in a safe place (it's secret, ha! ha!)"""
    ini_file = _get_app_data_location(DATA_FILE)
    if ini_file is None:
        return

    with open(ini_file, 'w') as configfile:
        config.write(configfile)



"""Dialog box to edit a 'section' object"""
from __future__ import annotations
import re

import tkinter as tk
import tkinter.font as tkFont
from tkinter import ttk
from tkinter.simpledialog import Dialog
from typing import Callable, TYPE_CHECKING

from .dialog_utilities import validate_class_times_equals_course_time, \
    get_block_info_from_tk_widgets, refresh_gui_blocks
from ..gui_generics.add_remove_tk import AddRemoveTk

if TYPE_CHECKING:
    from ..model import Stream, Lab, Teacher

# =====================================================================================================================
# Edit Section - add blocks, teachers, labs, streams
# =====================================================================================================================
class EditSectionDialogTk(Dialog):
    def __init__(self, frame: tk.Frame,
                 *,
                 course_description: str,
                 section_description: str = "",
                 assigned_teachers: list[Teacher] = None,
                 non_assigned_teachers: list[Teacher],
                 assigned_labs: list[Lab] = None,
                 non_assigned_labs: list[Lab],
                 assigned_streams: list[Stream] = None,
                 non_assigned_streams: list[Stream],
                 current_blocks: list[tuple[str, float, float]] = None,
                 course_hours: float,
                 apply_changes: Callable[[str, list, list, list, list], None]):
        """
        Edit Section - add blocks, teachers, labs, streams
        :param frame: the parent of the dialog box
        :param course_description: Number and name of course
        :param section_description: Description for Section
        :param assigned_teachers: which teachers have been assigned to this section
        :param non_assigned_teachers: which teachers have NOT been assigned to this section
        :param assigned_labs: which labs have been assigned to this section
        :param non_assigned_labs: which labs have not been assigned to this section
        :param assigned_streams: streams which have been assigned to this section
        :param non_assigned_streams: streams which have not been assigned to this section
        :param current_blocks: What blocks (class times) have been assigned
        :param apply_changes: what to call when the user is finished updating their request
        :param course_hours:
        """

        self.block_tk_variables = []
        self.course_hours = course_hours
        self.current_blocks = current_blocks if current_blocks is not None else []
        self.block_frames = None
        self.course_description = course_description
        self.top_frame = frame
        self._assigned_teachers = assigned_teachers if assigned_teachers is not None else []
        self._assigned_labs = assigned_labs if assigned_labs is not None else []
        self._assigned_streams = assigned_streams if assigned_streams is not None else []
        self._non_assigned_teachers = non_assigned_teachers
        self._non_assigned_labs = non_assigned_labs
        self._non_assigned_streams = non_assigned_streams
        self._apply_changes = apply_changes

        self.description = tk.StringVar(value=section_description)

        self.style = ttk.Style(frame.winfo_toplevel())

        dialog_title = "Edit Section"
        super().__init__(frame.winfo_toplevel(), dialog_title)

    # ================================================================================================================
    # The content of the main body of the dialog box
    # ================================================================================================================
    def body(self, frame: tk.Frame):

        # ------------------------------------------------------------------------------------------------------------
        # Info
        # ------------------------------------------------------------------------------------------------------------
        course_info_frame = tk.Frame(frame)
        section_info_frame = tk.Frame(frame)
        lbl = tk.Label(course_info_frame, text=self.course_description, anchor='center')
        lbl.pack(expand=1, fill='both', padx=15, pady=5)
        default_font = tkFont.nametofont(lbl.cget("font"))
        family = default_font["family"]
        size = default_font["size"] + 2
        lbl.config(font=(family, size))
        tk.Label(section_info_frame, text="Section Description", anchor='e', width=20).pack(side='left', padx=10,
                                                                                            pady=5)
        description = tk.Entry(section_info_frame, textvariable=self.description, )
        description.pack(side='left', padx=10, pady=5)

        # ------------------------------------------------------------------------------------------------------------
        # Blocks
        # ------------------------------------------------------------------------------------------------------------
        self.block_frames = tk.Frame(frame)
        for block_info in self.current_blocks:
            opt_day = tk.StringVar(value=block_info[0])
            opt_hour = tk.StringVar(value=str(block_info[1]))
            opt_duration = tk.StringVar(value=str(block_info[2]))
            self.block_tk_variables.append((opt_day, opt_hour, opt_duration))

        # ------------------------------------------------------------------------------------------------------------
        # Teacher/Lab/Stream Add/Remove
        # ------------------------------------------------------------------------------------------------------------
        teacher_assignments_frame = tk.Frame(frame)
        AddRemoveTk(teacher_assignments_frame, self._non_assigned_teachers, self._assigned_teachers,
                    "Assign Teacher to all Classes", "Remove Teacher from all Classes", height=6)

        lab_assignments_frame = tk.Frame(frame)
        AddRemoveTk(lab_assignments_frame, self._non_assigned_labs, self._assigned_labs,
                    "Assign Lab to all Classes", "Remove Lab from all Classes", height=6)

        stream_assignments_frame = tk.Frame(frame)
        AddRemoveTk(stream_assignments_frame, self._non_assigned_streams, self._assigned_streams,
                    "Assign Stream to Section", "Remove Stream from Section", height=6)

        # ------------------------------------------------------------------------------------------------------------
        # layout
        # ------------------------------------------------------------------------------------------------------------
        course_info_frame.grid(row=0, column=0)
        section_info_frame.grid(row=1, column=0)
        self.block_frames.grid(row=2, column=0, sticky='nsew', padx=5, pady=5)
        teacher_assignments_frame.grid(row=3, column=0, sticky='nsew', padx=5, pady=5)
        lab_assignments_frame.grid(row=4, column=0, sticky='nsew', padx=5, pady=5)
        stream_assignments_frame.grid(row=5, column=0, sticky='nsew', padx=5, pady=5)

        self.refresh()

        return description

    # ================================================================================================================
    # add a new block
    # ================================================================================================================
    def add_new_block(self):
        self.block_tk_variables.append((tk.StringVar(value="Monday"), tk.StringVar(value="8.0"), tk.StringVar(value="1.5")))
        self.refresh()

    # ================================================================================================================
    # update the gui to reflect the number of blocks
    # ================================================================================================================
    def refresh(self):
        refresh_gui_blocks(self)
        tk.Button(self.block_frames, text="Add New Class Time", command=self.add_new_block, padx=5).pack(expand=1, fill='y')

    # ================================================================================================================
    # are blocks set up properly?
    # ================================================================================================================
    def validate(self):
        if not validate_class_times_equals_course_time(self.block_tk_variables, float(self.course_hours)):
            return False
        return True

    # ================================================================================================================
    # apply changes
    # ================================================================================================================
    def apply(self):
        new_blocks = get_block_info_from_tk_widgets(self.block_tk_variables)
        self._apply_changes(self.description.get(), self._assigned_teachers, self._assigned_labs,
                            self._assigned_streams, new_blocks)



from typing import Any, Optional

import pytest

from src.scheduling_and_allocation.gui_pages import EditCoursesTk
from src.scheduling_and_allocation.model import Schedule, SemesterType, WeekDay
from src.scheduling_and_allocation.presenter.edit_courses import EditCourses


class EditCoursesTkTest(EditCoursesTk):
        # ========================================================================
        # constructor
        # ========================================================================
        def __init__(self, frame):
            self.structure = dict()
            self.called_clear_tree_text = False
            self.called_update_tree_text_arguments = []
            self.called_update_tree_text = False
            self.called_remove_tree_item_children = False
            self.called_remove_tree_item_children_arguments = []
            self.called_remove_tree_item = False
            self.called_remove_tree_item_arguments = []
            self.called_add_tree_item = False
            self.called_add_tree_item_arguments = []
            self.called_update_resource_type = False
            self.called_update_resource_type_arguments = []
            self.called_clear_tree = False
            self.frame = frame

        def get_new_id(self, parent_id):
            # find all current ids with this parent, and then add 1
            kid_ids = [iid for iid in self.structure.keys() if
                       iid.startswith(parent_id) and len(parent_id) < len(iid) < len(parent_id) + 5]
            id_numbers = [int(iid[-3:]) for iid in kid_ids]

            if len(id_numbers) == 0:
                id_numbers = [0]
            new_id_number = max(id_numbers) + 1

            return f"{parent_id}.{new_id_number:03d}"

        def update_resource_type_objects(self, resource_type, objs):
            self.called_update_resource_type = True
            self.called_update_resource_type_arguments.append ([resource_type])

        def add_tree_item(self, parent_id: str, name: str, child: Any, hide: bool = True):
            self.called_add_tree_item = True
            self.called_add_tree_item_arguments.append( [parent_id, name, child, hide])
            iid = self.get_new_id(parent_id)
            self.structure[iid] = name
            return iid

        def remove_tree_item(self, tree_iid: str):
            self.called_remove_tree_item = True
            self.called_remove_tree_item_arguments.append( [tree_iid])
            if tree_iid in self.structure.keys():
                self.structure.pop(tree_iid)
            else:
                raise ValueError(f"could not remove {tree_iid} from tree because it doesn't exist")

        def remove_tree_item_children(self, tree_iid: str):
            self.called_remove_tree_item_children = True
            self.called_remove_tree_item_children_arguments.append([tree_iid])

            ids = list(self.structure.keys())
            for this_iid in ids:
                if this_iid.startswith(tree_iid) and len(this_iid) > len(tree_iid):
                    self.structure.pop(this_iid)


        def update_tree_text(self, tree_iid: str, text: str):
            self.called_update_tree_text = True
            self.called_update_tree_text_arguments.append([tree_iid, text])
            if tree_iid in self.structure.keys():
                self.structure[tree_iid] = text

        def get_iid_from_name(self, name) -> str:
            for iid, s_name in self.structure.items():
                if name == s_name:
                    return iid
            return ""

        def clear_tree(self):
            self.called_clear_tree_text = True
            self.structure.clear()

        def get_parent(self, tree_id):
            return tree_id[0:-4]

        def get_iid_from_name_with_parent_id(self, name, parent_id) -> str:
            for iid, s_name in self.structure.items():
                s_name = s_name.replace("teacher: ","")
                s_name = s_name.replace("lab: ","")
                if name == s_name and iid.startswith(parent_id):
                    return iid
            return ""



@pytest.fixture()
def valid_tree_structure():
    return {'.001': '001 BasketWeaving',
        '.001.001': 'Section 1  (1A calculus available)',
        '.001.001.001': 'Class Time: Monday: 8:00 (1.5 hrs)',
        '.001.001.001.001': 'teacher: Jane Doe',
        '.001.001.001.002': 'lab: P107: C-Lab',
        '.001.001.002': 'Class Time: Monday: 10:00 (1.5 hrs)',
        '.001.001.002.001': 'teacher: Jane Doe',
        '.001.001.002.002': 'lab: P107: C-Lab',
        '.001.002': 'Section 2  (1B no calculus)',
        '.001.002.001': 'Class Time: Tuesday: 8:00 (1.5 hrs)',
        '.001.002.001.002': 'lab: P322: Mac Lab',
        '.001.002.001.001': 'teacher: John Doe',
        '.001.002.002': 'Class Time: Tuesday: 10:00 (1.5 hrs)',
        '.001.002.002.002': 'lab: P325',
        '.001.002.002.001': 'teacher: Babe Ruth',
        '.002': '002 Thumb Twiddling',
        '.002.001': 'Section 1',
        '.002.001.001': 'Class Time: Monday: 8:00 (1.5 hrs)',
        '.002.001.002': 'Class Time: Monday: 10:00 (1.5 hrs)',
        '.002.002': 'Section 2',
        '.002.002.001': 'Class Time: Tuesday: 8:00 (1.5 hrs)',
        '.002.002.002': 'Class Time: Tuesday: 10:00 (1.5 hrs)'}


@pytest.fixture()
def schedule_obj():
    schedule = Schedule()

    # teachers
    t1 = schedule.add_update_teacher("Jane", "Doe", "0.25", teacher_id="1")
    t2 = schedule.add_update_teacher("John", "Doe", teacher_id="2")
    t3 = schedule.add_update_teacher("Babe", "Ruth", teacher_id="3")
    t4 = schedule.add_update_teacher("Bugs", "Bunny", teacher_id = "4")

    # labs
    l1 = schedule.add_update_lab("P107", "C-Lab")
    l2 = schedule.add_update_lab("P322", "Mac Lab")
    l3 = schedule.add_update_lab("P325")
    l4 = schedule.add_update_lab("BH311", "Britain Hall")

    # streams
    st1 = schedule.add_update_stream("1A", "calculus available")
    st2 = schedule.add_update_stream("1B", "no calculus")
    st3 = schedule.add_update_stream("2A")
    st4 = schedule.add_update_stream("2B")

    # courses
    c_001 = schedule.add_update_course("001","BasketWeaving", SemesterType.fall )
    s_001_1 = c_001.add_section("1", section_id=1)
    s_001_1.add_stream(st1)
    b_001_1_1 = s_001_1.add_block(WeekDay.Monday, 8 )
    b_001_1_2 = s_001_1.add_block(WeekDay.Monday,  10 )
    b_001_1_1.add_teacher(t1)
    b_001_1_2.add_teacher(t1)
    b_001_1_1.add_lab(l1)
    b_001_1_2.add_lab(l1)

    s_001_2 = c_001.add_section("2",section_id=2)
    s_001_2.add_stream(st2)
    b_001_2_1 = s_001_2.add_block(WeekDay.Tuesday,  8 )
    b_001_2_2 = s_001_2.add_block(WeekDay.Tuesday,  10 )
    b_001_2_1.add_teacher(t2)
    b_001_2_2.add_teacher(t3)
    b_001_2_1.add_lab(l2)
    b_001_2_2.add_lab(l3)


    c_002 = schedule.add_update_course("002","Thumb Twiddling", SemesterType.fall )
    s_002_1 = c_002.add_section("1",)
    b_002_1_1 = s_002_1.add_block(WeekDay.Monday,  8 )
    b_002_1_2 = s_002_1.add_block(WeekDay.Monday,  10 )
    s_002_2 = c_002.add_section("2",)
    b_002_2_1 = s_002_2.add_block(WeekDay.Tuesday,  8 )
    b_002_2_2 = s_002_2.add_block(WeekDay.Tuesday,  10 )



    return schedule

@pytest.fixture()
def gui():
    return EditCoursesTkTest(None)

dirty_flag = False


@pytest.fixture()
def dirty():
    global dirty_flag
    dirty_flag = False

    def set_dirty(value: Optional[bool] = None) -> bool:
        global dirty_flag
        if value is not None:
            dirty_flag = value
        return dirty_flag

    return set_dirty

# =============================================================================
# Tree created properly
# =============================================================================
def test_init_edit_course(gui, dirty, schedule_obj):
    """can initialize an instance of EditCourses"""

    # execute
    ec = EditCourses(dirty, None, schedule_obj, gui)

    # verify
    assert ec is not None
    assert isinstance(ec, EditCourses)

def test_refresh_updates_tree(gui, dirty, schedule_obj, valid_tree_structure):
    """can create a pop-up menu for course and all required items are there"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)

    # execute
    ec.refresh()

    # verify
    for key in valid_tree_structure.keys():
        assert key in gui.structure
        assert gui.structure[key] == valid_tree_structure[key]



# =============================================================================
# Tests for LAB/TEACHER menu
# =============================================================================
def test_create_tree_menu_for_teacher(gui, dirty, schedule_obj):
    """can create a pop-up menu for teacher and all required items are there"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    teacher = block.teachers()[0]
    tree_id = ".001.001.001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(teacher, block, tree_id, parent_id)

    # verify
    assert len([mi for mi in menu if mi.label == 'Remove Teacher']) == 1

def test_create_tree_menu_for_teacher_remove_teacher(gui, dirty, schedule_obj):
    """remove teacher from parent"""
    # NOTE: teachers method returns sorted teachers by first name,
    # so  Babe Ruth comes before Jane Doe

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    block.add_teacher(schedule_obj.get_teacher_by_name("Babe","Ruth"))
    teacher = block.teachers()[0]
    ec.refresh()
    parent_id = ".001.001.001"
    tree_id = gui.get_iid_from_name_with_parent_id(f"{teacher.firstname} {teacher.lastname}", parent_id)


    menu = ec.tree_event_create_tree_popup(teacher, block, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove Teacher'][0]

    # execute
    num_teachers = len(block.teachers())
    mi.command()

    # verify
    assert len(block.teachers()) == num_teachers - 1
    x=block.teachers()
    for iid in [iid for iid in gui.structure.keys() if iid.startswith(parent_id)]:
        assert not gui.structure[iid].startswith('teacher: Babe Ruth')
    assert gui.get_iid_from_name_with_parent_id("Jane Doe", parent_id)


# =============================================================================
# Tests for LAB/TEACHER menu
# =============================================================================
def test_create_tree_menu_for_lab(gui, dirty, schedule_obj):
    """can create a pop-up menu for lab and all required items are there"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    lab = block.labs()[0]
    tree_id = ".001.001.001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(lab, block, tree_id, parent_id)

    # verify
    assert len([mi for mi in menu if mi.label == 'Remove Lab']) == 1

def test_create_tree_menu_for_lab_remove_lab(gui, dirty, schedule_obj, valid_tree_structure):
    """remove lab from parent"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    ec.refresh()
    block.add_lab(schedule_obj.get_lab_by_number("P325"))
    lab = block.labs()[0]
    tree_id = gui.get_iid_from_name_with_parent_id(lab.number, ".001.001.001")
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(lab, block, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove Lab'][0]


    # execute
    num_labs = len(block.labs())
    mi.command()

    # verify
    assert len(block.labs()) == num_labs - 1
    for iid in [iid for iid in gui.structure.keys() if iid.startswith(parent_id)]:
        assert not gui.structure[iid].startswith(lab.number)



# =============================================================================
# Tests for BLOCK menu
# =============================================================================

def test_create_tree_menu_for_block(gui, dirty, schedule_obj):
    """can create a pop-up menu for block and all required items are there"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    tree_id = ".001.001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(block, section, tree_id, parent_id)

    # verify
    assert len([mi for mi in menu if mi.label == 'Remove Class Time']) == 1
    assert len([mi for mi in menu if mi.label == 'Add Teacher']) == 1
    assert len([mi for mi in menu if mi.label == 'Add Lab']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove Teacher']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove Lab']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove All']) == 1


def test_tree_menu_for_section_remove_block(gui, dirty, schedule_obj):
    """run menu command for block: remove
    - dirty flag is set
    - tree is cleared of block
    - schedule has been updated
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    tree_id = ".001.001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(block, section, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove Class Time'][0]
    assert gui.get_iid_from_name_with_parent_id("Class Time: Monday: 8:00 (1.5 hrs)", parent_id)

    # execute
    mi.command()

    # verify
    assert dirty_flag
    assert len(section.blocks()) == 1
    assert not gui.get_iid_from_name_with_parent_id("Class Time: Monday: 8:00 (1.5 hrs)", parent_id)

def test_tree_menu_for_block_add_teacher_sub_menu(gui, dirty, schedule_obj):
    """run menu command for block: add teachers
    - sub menu only contains teachers not already assigned to block
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    tree_id = ".001.001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(block, section, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == 'Add Teacher'][0]
    assert len(mi.children) == 3
    assert mi.children[0].label == "Babe Ruth"
    assert mi.children[1].label == "Bugs Bunny"
    assert mi.children[2].label == "John Doe"

def test_tree_menu_for_block_add_teacher(gui, dirty, schedule_obj):
    """run menu command for block: add teachers - select teacher
    - teacher added to block/schedule (but only to this block, not the other one)
    - teacher added to block/tree
    - dirty flag is set
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    block2 = section.blocks()[1]
    tree_id = ".001.001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(block, section, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == "Add Teacher"]
    add_menu = mi[0].children[1]        # Bugs Bunny

    # execute
    add_menu.command()

    # verify
    assert dirty_flag
    assert block.has_teacher(schedule_obj.get_teacher_by_name("Bugs", "Bunny"))
    assert not block2.has_teacher(schedule_obj.get_teacher_by_name("Bugs", "Bunny"))
    assert gui.get_iid_from_name_with_parent_id("Bugs Bunny", tree_id)
    new_iid = gui.get_iid_from_name_with_parent_id("Bugs Bunny", tree_id)
    assert len(new_iid) == 4*4, "course_id + section_id + block_id + teacher_id"
    assert not gui.get_iid_from_name_with_parent_id("John Doe", ".001.001.002")


def test_tree_menu_for_block_add_lab_sub_menu(gui, dirty, schedule_obj):
    """run menu command for block: add labs
    - sub menu only contains labs not already assigned to block
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    tree_id = ".001.001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(block, section, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == "Add Lab"]

    assert len(mi[0].children) == 3
    assert mi[0].children[0].label == "BH311: Britain Hall"
    assert mi[0].children[1].label == "P322: Mac Lab"
    assert mi[0].children[2].label == "P325"

def test_tree_menu_for_block_add_lab(gui, dirty, schedule_obj):
    """run menu command for block: add lab - select lab
    - lab added to section/schedule
    - lab added to section/tree
    - dirty flag is set
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    tree_id = ".001.001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(block, section, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == "Add Lab"]
    add_menu = mi[0].children[0]

    # execute
    add_menu.command()

    # verify
    assert dirty_flag
    assert section.has_lab(schedule_obj.get_lab_by_number("BH311"))
    assert gui.get_iid_from_name_with_parent_id("BH311: Britain Hall", tree_id)
    new_iid = gui.get_iid_from_name_with_parent_id("BH311: Britain Hall", tree_id)
    assert len(new_iid) == 4*4, "course_id + section_id + block_id + teacher_id"
    assert not gui.get_iid_from_name_with_parent_id("BH311: Britain Hall", ".001.001.002")


def test_tree_menu_for_block_remove_all_sub_menu(gui, dirty, schedule_obj):
    """create remove_all sub menu
    - sub menu for sections, streams, labs, and teachers"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    tree_id = ".001.001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(block, section, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children
    assert len(sub_menu) == 2
    assert sub_menu[0].label == "labs"
    assert sub_menu[1].label == "teachers"


def test_tree_menu_for_blocks_remove_all_labs(gui, dirty, schedule_obj):
    """run menu command for block: remove_all labs
    - dirty flag is set
    - all labs, etc have been removed from block/tree
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    tree_id = ".001.001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(block, section, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'labs'][0]
    mi2.command()

    assert dirty_flag
    assert len(block.labs()) == 0
    ids = [iid for iid in gui.structure.keys() if iid.startswith(tree_id)]
    for iid in ids:
        assert not gui.structure[iid].startswith("P107")
        assert not gui.structure[iid].startswith("P322")
        assert not gui.structure[iid].startswith("P325")


def test_tree_menu_for_block_remove_all_teachers(gui, dirty, schedule_obj):
    """run menu command for section: remove_all teacher
    - dirty flag is set
    - all teachers, etc have been removed from block/tree
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    block = section.blocks()[0]
    tree_id = ".001.001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(block, section, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'teachers'][0]
    mi2.command()

    assert dirty_flag
    assert len(block.teachers()) == 0
    ids = [iid for iid in gui.structure.keys() if iid.startswith(tree_id)]
    for iid in ids:
        assert not gui.structure[iid].startswith("Jane Doe")
        assert not gui.structure[iid].startswith("John Doe")
        assert not gui.structure[iid].startswith("Babe Ruth")
        assert not gui.structure[iid].startswith("Bugs Bunny")


# =============================================================================
# Tests for SECTION menu
# =============================================================================


def test_create_tree_menu_for_section(gui, dirty, schedule_obj):
    """can create a pop-up menu for section and all required items are there"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)

    # verify
    assert len([mi for mi in menu if mi.label == 'Edit Section']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove Section']) == 1
    assert len([mi for mi in menu if mi.label == 'Add Class Times']) == 1
    assert len([mi for mi in menu if mi.label == 'Add Teacher']) == 1
    assert len([mi for mi in menu if mi.label == 'Add Lab']) == 1
    assert len([mi for mi in menu if mi.label == 'Add Stream']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove Teacher']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove Lab']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove Stream']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove All']) == 1

def test_tree_menu_for_section_edit_section(gui, dirty, schedule_obj):
    """click edit section
    - EditCourses.edit_section_dialog must be called"""

    called_edit_dialog = False
    def edit_dialog_test(section, tree_id):
        nonlocal called_edit_dialog
        called_edit_dialog = True


    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)

    ec.edit_section_dialog = edit_dialog_test
    mi = [mi for mi in menu if mi.label == 'Edit Section'][0]

    # execute
    mi.command()

    # verify
    assert called_edit_dialog

def test_tree_menu_for_section_remove_section(gui, dirty, schedule_obj):
    """run menu command for section: remove
    - dirty flag is set
    - tree is cleared of section
    - schedule has been updated
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove Section'][0]
    assert gui.get_iid_from_name_with_parent_id("Section 1  (1A calculus available)", ".001")

    # execute
    mi.command()

    # verify
    assert dirty_flag
    assert len(course.sections()) == 1
    assert not gui.get_iid_from_name_with_parent_id("Section 1  (1A calculus available)", ".001")

def test_tree_menu_for_section_add_blocks(gui, dirty, schedule_obj):
    """click section/add blocks
    - EditCourses.add_blocks_dialog must be called"""

    called_add_dialog = False
    def add_dialog_test(course, tree_id):
        nonlocal called_add_dialog
        called_add_dialog = True


    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)
    ec.add_blocks_dialog = add_dialog_test
    mi = [mi for mi in menu if mi.label == 'Add Class Times'][0]

    # execute
    mi.command()

    # verify
    assert called_add_dialog

def test_tree_menu_for_section_add_teacher_sub_menu(gui, dirty, schedule_obj):
    """run menu command for section: add teachers
    - sub menu only contains teachers not already assigned to section
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == 'Add Teacher'][0]
    assert len(mi.children) == 3
    assert mi.children[0].label == "Babe Ruth"
    assert mi.children[1].label == "Bugs Bunny"
    assert mi.children[2].label == "John Doe"

def test_tree_menu_for_section_add_teacher(gui, dirty, schedule_obj):
    """run menu command for section: add teachers - select teacher
    - teacher added to section/schedule
    - teacher added to section/tree in correct spot
    - dirty flag is set
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == "Add Teacher"]
    add_menu = mi[0].children[1]        # Bugs Bunny

    # execute
    add_menu.command()

    # verify
    assert dirty_flag
    assert section.has_teacher(schedule_obj.get_teacher_by_name("Bugs", "Bunny"))
    assert gui.get_iid_from_name_with_parent_id("Bugs Bunny", tree_id)
    new_iid = gui.get_iid_from_name_with_parent_id("Bugs Bunny", tree_id)
    assert len(new_iid) == 4*4, "course_id + section_id + block_id + teacher_id"


def test_tree_menu_for_section_add_stream_sub_menu(gui, dirty, schedule_obj):
    """run menu command for section: add streams
    - sub menu only contains streams not already assigned to section
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == "Add Stream"]
    assert len(mi[0].children) == 3
    assert mi[0].children[0].label == "1B no calculus"
    assert mi[0].children[1].label == "2A "
    assert mi[0].children[2].label == "2B "

def test_tree_menu_for_section_add_stream(gui, dirty, schedule_obj):
    """run menu command for section: add stream - select stream
    - stream added to section/schedule
    - stream added to section/tree
    - dirty flag is set
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == "Add Stream"]
    add_menu = mi[0].children[1]

    # execute
    add_menu.command()

    # verify
    assert dirty_flag
    assert section.has_stream(schedule_obj.get_stream_by_number("2A"))

def test_tree_menu_for_section_add_lab_sub_menu(gui, dirty, schedule_obj):
    """run menu command for section: add labs
    - sub menu only contains labs not already assigned to course
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == "Add Lab"]

    assert len(mi[0].children) == 3
    assert mi[0].children[0].label == "BH311: Britain Hall"
    assert mi[0].children[1].label == "P322: Mac Lab"
    assert mi[0].children[2].label == "P325"

def test_tree_menu_for_section_add_lab(gui, dirty, schedule_obj):
    """run menu command for section: add lab - select lab
    - lab added to section/schedule
    - lab added to section/tree
    - dirty flag is set
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == "Add Lab"]
    add_menu = mi[0].children[0]

    # execute
    add_menu.command()

    # verify
    assert dirty_flag
    assert section.has_lab(schedule_obj.get_lab_by_number("BH311"))
    assert gui.get_iid_from_name_with_parent_id("BH311: Britain Hall", tree_id)

    new_iid = gui.get_iid_from_name_with_parent_id("BH311: Britain Hall", tree_id)
    assert len(new_iid) == 4*4, "course_id + section_id + block_id + lab_id"

def test_tree_menu_for_section_remove_all_sub_menu(gui, dirty, schedule_obj):
    """create remove_all sub menu
    - sub menu for sections, streams, labs, and teachers"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children
    assert len(sub_menu) == 4
    assert sub_menu[0].label == "class times"
    assert sub_menu[1].label == "streams"
    assert sub_menu[2].label == "labs"
    assert sub_menu[3].label == "teachers"


def test_tree_menu_for_section_remove_all_blocks(gui, dirty, schedule_obj):
    """run menu command for section: remove_all block
    - dirty flag is set
    - all blocks, etc have been removed from section/tree
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'class times'][0]
    mi2.command()

    assert dirty_flag
    assert len(section.blocks()) == 0
    assert ".001.001.001" not in gui.structure.keys()
    assert ".001.001.002" not in gui.structure.keys()

def test_tree_menu_for_section_remove_all_streams(gui, dirty, schedule_obj):
    """run menu command for section: remove_all streams
    - dirty flag is set
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'streams'][0]
    mi2.command()

    assert dirty_flag
    assert len(section.streams()) == 0

def test_tree_menu_for_section_remove_all_labs(gui, dirty, schedule_obj):
    """run menu command for section: remove_all section
    - dirty flag is set
    - all labs, etc have been removed from section/tree
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'labs'][0]
    mi2.command()

    assert dirty_flag
    assert len(section.labs()) == 0
    ids = [iid for iid in gui.structure.keys() if iid.startswith(tree_id)]
    for iid in ids:
        assert not gui.structure[iid].startswith("P107")
        assert not gui.structure[iid].startswith("P322")
        assert not gui.structure[iid].startswith("P325")


def test_tree_menu_for_section_remove_all_teachers(gui, dirty, schedule_obj):
    """run menu command for section: remove_all teacher
    - dirty flag is set
    - all teachers, etc have been removed from section/tree
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    course = schedule_obj.get_course_by_number("001")
    section = course.get_section_by_id(1)
    tree_id = ".001.001"
    parent_id = gui.get_parent(tree_id)
    menu = ec.tree_event_create_tree_popup(section, course, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'teachers'][0]
    mi2.command()

    assert dirty_flag
    assert len(section.teachers()) == 0
    ids = [iid for iid in gui.structure.keys() if iid.startswith(tree_id)]
    for iid in ids:
        assert not gui.structure[iid].startswith("Jane Doe")
        assert not gui.structure[iid].startswith("John Doe")
        assert not gui.structure[iid].startswith("Babe Ruth")
        assert not gui.structure[iid].startswith("Bugs Bunny")

# =============================================================================
# Tests for COURSE menu
# =============================================================================
def test_create_tree_menu_for_course(gui, dirty, schedule_obj):
    """can create a pop-up menu for course and all required items are there"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)

    # verify
    assert len([mi for mi in menu if mi.label == 'Edit Course']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove Course']) == 1
    assert len([mi for mi in menu if 'needs allocation' in mi.label]) == 1
    assert len([mi for mi in menu if mi.label == 'Add Sections']) == 1
    assert len([mi for mi in menu if mi.label == 'Add Teacher']) == 1
    assert len([mi for mi in menu if mi.label == 'Add Lab']) == 1
    assert len([mi for mi in menu if mi.label == 'Add Stream']) == 1
    assert len([mi for mi in menu if mi.label == 'Remove All']) == 1

def test_tree_menu_for_course_edit_course(gui, dirty, schedule_obj):
    """click edit course
    - EditCourses.edit_course_dialog must be called"""

    called_edit_dialog = False
    def edit_course_dialog_test(course, tree_id):
        nonlocal called_edit_dialog
        called_edit_dialog = True


    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, "001", "")
    ec.edit_course_dialog = edit_course_dialog_test
    mi = [mi for mi in menu if mi.label == 'Edit Course'][0]

    # execute
    mi.command()

    # verify
    assert called_edit_dialog

def test_tree_menu_for_course_remove_course(gui, dirty, schedule_obj):
    """run menu command for course: remove_course
    - dirty flag is set
    - tree is cleared of course
    - schedule has been update
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)

    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove Course'][0]

    # execute
    mi.command()

    # verify
    assert dirty_flag
    assert schedule_obj.get_course_by_number("001") is None
    assert schedule_obj.get_course_by_number("002") is not None
    assert gui.clear_tree
    assert not gui.get_iid_from_name("001 BasketWeaving")
    assert gui.get_iid_from_name("002 Thumb Twiddling")

def test_tree_menu_for_course_unset_needs_allocation(gui, dirty, schedule_obj):
    """run menu command for course: needs allocation
    - dirty flag is set
    - needs allocation is unset
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)

    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if "allocation" in mi.label and 'Unset' in mi.label][0]

    # execute
    mi.command()

    # verify
    assert dirty_flag
    assert not schedule_obj.get_course_by_number("001").needs_allocation

def test_tree_menu_for_course_set_needs_allocation(gui, dirty, schedule_obj):
    """run menu command for course: needs allocation
    - dirty flag is set
    - needs allocation is unset
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)
    course = schedule_obj.get_course_by_number("001")
    course.needs_allocation = False
    assert not schedule_obj.get_course_by_number("001").needs_allocation

    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if "allocation" in mi.label and 'Set' in mi.label][0]

    # execute
    mi.command()

    # verify
    assert dirty_flag
    assert schedule_obj.get_course_by_number("001").needs_allocation

def test_tree_menu_for_course_add_sections(gui, dirty, schedule_obj):
    """click course/add sections
    - EditCourses.add_section_dialog must be called"""

    called_add_section_dialog = False
    def add_section_dialog_test(course, tree_id):
        nonlocal called_add_section_dialog
        called_add_section_dialog = True


    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, "001", "")
    ec.add_section_dialog = add_section_dialog_test
    mi = [mi for mi in menu if mi.label == 'Add Sections'][0]

    # execute
    mi.command()

    # verify
    assert called_add_section_dialog

def test_tree_menu_for_course_add_teacher_sub_menu(gui, dirty, schedule_obj):
    """run menu command for course: add teachers
    - sub menu only contains teachers not already assigned to course
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)
    schedule_obj.add_update_teacher("Mickey","Mouse")

    # execute
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == "Add Teacher"]
    assert len(mi[0].children) == 2
    assert mi[0].children[0].label == "Bugs Bunny"
    assert mi[0].children[1].label == "Mickey Mouse"

def test_tree_menu_for_course_add_teacher(gui, dirty, schedule_obj):
    """run menu command for course: add teachers - select teacher
    - teacher added to course/schedule
    - teacher added to course/tree
    - dirty flag is set
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)
    schedule_obj.add_update_teacher("Mickey","Mouse")
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == "Add Teacher"]
    add_menu = mi[0].children[1]

    # execute
    add_menu.command()

    # verify
    assert dirty_flag
    assert schedule_obj.get_course_by_number("001").has_teacher(schedule_obj.get_teacher_by_name("Mickey", "Mouse"))
    assert gui.get_iid_from_name_with_parent_id("Mickey Mouse", tree_id)
    new_iid = gui.get_iid_from_name_with_parent_id("Mickey Mouse", tree_id)
    assert len(new_iid) == 4*4, "course_id + section_id + block_id + teacher_id"

def test_tree_menu_for_course_add_stream_sub_menu(gui, dirty, schedule_obj):
    """run menu command for course: add streams
    - sub menu only contains streams not already assigned to course
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == "Add Stream"]
    assert len(mi[0].children) == 2
    assert mi[0].children[0].label == "2A "
    assert mi[0].children[1].label == "2B "

def test_tree_menu_for_course_add_stream(gui, dirty, schedule_obj):
    """run menu command for course: add stream - select stream
    - stream added to course/schedule
    - stream added to course/tree
    - dirty flag is set
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == "Add Stream"]
    add_menu = mi[0].children[1]

    # execute
    add_menu.command()

    # verify
    assert dirty_flag
    assert schedule_obj.get_course_by_number("001").has_stream(schedule_obj.get_stream_by_number("2B"))

def test_tree_menu_for_course_add_lab_sub_menu(gui, dirty, schedule_obj):
    """run menu command for course: add labs
    - sub menu only contains labs not already assigned to course
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == "Add Lab"]

    assert len(mi[0].children) == 1
    assert mi[0].children[0].label == "BH311: Britain Hall"

def test_tree_menu_for_course_add_lab(gui, dirty, schedule_obj):
    """run menu command for course: add lab - select lab
    - lab added to course/schedule
    - lab added to course/tree
    - dirty flag is set
    """

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == "Add Lab"]
    add_menu = mi[0].children[0]

    # execute
    add_menu.command()

    # verify
    assert dirty_flag
    assert schedule_obj.get_course_by_number("001").has_lab(schedule_obj.get_lab_by_number("BH311"))
    assert gui.get_iid_from_name_with_parent_id("BH311: Britain Hall", parent_id)
    new_iid = gui.get_iid_from_name_with_parent_id("BH311: Britain Hall", tree_id)
    assert len(new_iid) == 4*4, "course_id + section_id + block_id + lab_id"

def test_tree_menu_for_course_remove_all_sub_menu(gui, dirty, schedule_obj):
    """create remove_all sub menu
    - sub menu for sections, streams, labs, and teachers"""

    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)

    # execute
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)

    # verify
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children
    assert len(sub_menu) == 4
    assert sub_menu[0].label == "sections"
    assert sub_menu[1].label == "streams"
    assert sub_menu[2].label == "labs"
    assert sub_menu[3].label == "teachers"


def test_tree_menu_for_course_remove_all_sections(gui, dirty, schedule_obj):
    """run menu command for course: remove_all section
    - dirty flag is set
    - all sections, etc have been removed from course/tree
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'sections'][0]
    mi2.command()

    assert dirty_flag
    course = schedule_obj.get_course_by_number("001")
    assert len(course.sections()) == 0
    assert ".001.001" not in gui.structure.keys()
    assert ".001.002" not in gui.structure.keys()

def test_tree_menu_for_course_remove_all_streams(gui, dirty, schedule_obj):
    """run menu command for course: remove_all streams
    - dirty flag is set
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'streams'][0]
    mi2.command()

    assert dirty_flag
    course = schedule_obj.get_course_by_number("001")
    assert len(course.streams()) == 0

def test_tree_menu_for_course_remove_all_labs(gui, dirty, schedule_obj):
    """run menu command for course: remove_all lab
    - dirty flag is set
    - all labs, etc have been removed from course/tree
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'labs'][0]
    mi2.command()

    assert dirty_flag
    course = schedule_obj.get_course_by_number("001")
    assert len(course.labs()) == 0
    ids = [iid for iid in gui.structure.keys() if iid.startswith(parent_id)]
    for iid in ids:
        assert not gui.structure[iid].startswith("P107")
        assert not gui.structure[iid].startswith("P322")
        assert not gui.structure[iid].startswith("P325")

def test_tree_menu_for_course_remove_all_teachers(gui, dirty, schedule_obj):
    """run menu command for course: remove_all teacher
    - dirty flag is set
    - all teachers, etc have been removed from course/tree
    - schedule has been updated
    """
    # prepare
    ec = EditCourses(dirty, None, schedule_obj, gui)
    ec.refresh()
    tree_id = gui.get_iid_from_name("001 BasketWeaving")
    parent_id = gui.get_iid_from_name(tree_id)
    menu = ec.tree_event_create_tree_popup(schedule_obj.get_course_by_number("001"), None, tree_id, parent_id)
    mi = [mi for mi in menu if mi.label == 'Remove All'][0]
    sub_menu = mi.children

    # execute
    mi2 = [mi for mi in sub_menu if mi.label == 'teachers'][0]
    mi2.command()

    assert dirty_flag
    course = schedule_obj.get_course_by_number("001")
    assert len(course.teachers()) == 0
    ids = [iid for iid in gui.structure.keys() if iid.startswith(tree_id)]
    for iid in ids:
        assert not gui.structure[iid].startswith("Jane Doe")
        assert not gui.structure[iid].startswith("John Doe")
        assert not gui.structure[iid].startswith("Babe Ruth")
        assert not gui.structure[iid].startswith("Bugs Bunny")






"""
# ============================================================================
# Code that draws the View stuff.
#
# PURPOSE: Provides all the gui interactions with the View
#
#
# EVENT HANDLERS:
#         get_popup_menu_handler(gui_block_id)
#         refresh_blocks_handler()
#         on_closing_handler()
#         double_click_block_handler(gui_id)
#         gui_block_is_moving_handler(gui_id, day, start)
#         gui_block_has_dropped_handler(gui_id)
#         undo_handler()
#         redo_handler()
# ============================================================================
"""
from __future__ import annotations

from functools import partial
import tkinter as tk
from typing import Callable

from ..gui_generics.menu_and_toolbars import generate_menu, MenuItem, MenuType
from ..Utilities import Colour
from ..Utilities.id_generator import IdGenerator
from ..gui_generics.block_colours import get_conflict_colour_info, IMMOVABLE_COLOUR, RESOURCE_COLOURS
from ..gui_pages.view_canvas_tk import ViewCanvasTk
from ..model import ResourceType, ConflictType

DEFAULT_CANVAS_WIDTH = 700
DEFAULT_CANVAS_HEIGHT = 700


def _default_menu(*_) -> list[MenuItem]:
    menu = MenuItem(name="nothing", label="nothing", menu_type=MenuType.Command, command=lambda: None)
    return [menu, ]


class ViewDynamicTk:
    """ Provide the dynamic view for all calls times for a specific resource."""

    view_ids = IdGenerator()

    # =================================================================================================================
    # Init
    # =================================================================================================================
    def __init__(self, frame: tk.Frame, title: str, resource_type: ResourceType,):
        """
        creates a dynamic view to see all the blocks, where they are, movable, changeable, etc.
        :param frame: where to put the view
        :param title: title of the toplevel window (displayed in the top bar)
        :param resource_type: What type of view is this
        """

        self.mw = frame.winfo_toplevel()
        self.view_id = self.view_ids.get_new_id()

        # ------------------------------------------------------------------------------------------------------------
        # handlers
        # ------------------------------------------------------------------------------------------------------------
        # returns menu info for gui block popup menu
        self.get_popup_menu_handler: Callable[[str], list[MenuItem]] = _default_menu

        # redraws all the appropriate blocks on the canvas
        self.refresh_blocks_handler: Callable[[], None] = lambda: None

        # handles the closing of this view
        self.on_closing_handler: Callable = lambda *_: None

        # handles a double click on a gui block
        self.double_click_block_handler: Callable[[str], None] = lambda gui_id: None

        # handles the movement of a gui block
        self.gui_block_is_moving_handler: Callable[[str, float, float], None] = lambda gui_id, day, start: None

        # handles the dropping of a gui block
        self.gui_block_has_dropped_handler: Callable[[str], None] = lambda gui_id: None

        # undo/redo handlers
        self.undo_handler: Callable[[], None] = lambda: None
        self.redo_handler: Callable[[], None] = lambda: None

        # ------------------------------------------------------------------------------------------------------------
        # create a new toplevel window for this view
        # ------------------------------------------------------------------------------------------------------------
        tl = tk.Toplevel(self.mw)
        self.toplevel = tl
        tl.protocol('WM_DELETE_WINDOW', self.destroy)
        tl.resizable(False, False)
        tl.title(title)

        # ------------------------------------------------------------------------------------------------------------
        # Create bar at top to show Colour coding of conflicts
        # ------------------------------------------------------------------------------------------------------------
        f = tk.Frame(tl)
        f.pack(expand=1, fill="x")

        for c in get_conflict_colour_info(resource_type):
            label_text = c["text"]
            label_text = label_text.replace("_"," ")
            label_text = label_text.capitalize()
            tk.Label(f, text=label_text, width=10, background=c['bg'], foreground=c['fg']) \
                .pack(side='left', expand=1, fill="x")

        # ------------------------------------------------------------------------------------------------------------
        # add canvas
        # ------------------------------------------------------------------------------------------------------------
        cn = tk.Canvas(tl, height=DEFAULT_CANVAS_HEIGHT, width=DEFAULT_CANVAS_WIDTH, background="white")
        cn.pack()
        self.cn = cn

        # create the view_canvas
        self.view_canvas = self._draw_view_canvas()

        # ------------------------------------------------------------------------------------------------------------
        # create scale menu and redo/undo menu
        # ------------------------------------------------------------------------------------------------------------
        main_menu = tk.Menu(tl)
        tl.configure(menu=main_menu)

        scale_menu = MenuItem(name='view', menu_type=MenuType.Cascade, label='View')
        scale_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                      label='50%', command=partial(self.draw, 0.50) )
                             )
        scale_menu.add_child(MenuItem( menu_type=MenuType.Command,
                                      label='75%', command=partial(self.draw, 0.75))
                             )
        scale_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                      label='100%', command=partial(self.draw, 1.0))
                             )
        undo_menu = MenuItem(name='undo_redo', menu_type=MenuType.Cascade, label='Undo/Redo')
        undo_menu.add_child(MenuItem( menu_type=MenuType.Command, label='Undo',
                                     accelerator="Control-z", command=self._undo )
                            )
        undo_menu.add_child(MenuItem( menu_type=MenuType.Command, label='Redo',
                                     accelerator="Control-y", command=self._redo )
                            )

        # return list of top level menu items
        generate_menu(tl, [scale_menu, undo_menu], main_menu)

    # =================================================================================================================
    # draw the view canvas (the static drawing of the view)
    # =================================================================================================================
    def _draw_view_canvas(self, scale_factor: float = 1.0) -> ViewCanvasTk:
        """
        Clears the canvas of all blocks and redraws the background, followed by a call
        to an event handler that will update all the required blocks
        :param scale_factor: Allows smaller views to be drawn, 1=100%
        """

        self.cn.delete('all')

        # remove any binding to the canvas itself.
        self.cn.bind("<1>", "")
        self.cn.bind("<B1-Motion>", "")
        self.cn.bind("<ButtonRelease-1>", "")

        # set height and width of canvas and toplevel
        self.cn.configure(width=DEFAULT_CANVAS_WIDTH * scale_factor,
                          height=DEFAULT_CANVAS_WIDTH * scale_factor)

        # redraw background (things that don't change, like time, etc.)
        self.view_canvas = ViewCanvasTk(self.cn, scale_factor)

        # update size of window
        self.toplevel.update_idletasks()  # Update widget layout and dimensions
        width = self.toplevel.winfo_reqwidth()  # Get required width based on content
        height = self.toplevel.winfo_reqheight()  # Get required height based on content
        self.toplevel.geometry(f"{width}x{height}")

        return self.view_canvas

    # =================================================================================================================
    # draw the view (the view with all the gui interactions) (called by view, and if resized from menu)
    # =================================================================================================================
    def get_scale(self):
        return self.view_canvas.scale.current_scale

    def draw(self, scale_factor: float = 1.0):
        """
        Draw the view, calls event handler refresh_blocks_handler to get info about the blocks
        :param scale_factor:
        :return:
        """
        self._draw_view_canvas(scale_factor)

        # update blocks
        self.refresh_blocks_handler()

        # reset binding to canvas objects
        self.cn.tag_bind(self.view_canvas.Movable_Tag_Name, "<Button-1>", self._select_gui_block_to_move)
        self.cn.tag_bind(self.view_canvas.Clickable_Tag_Name, '<3>', self._post_menu)
        self.cn.tag_bind(self.view_canvas.Clickable_Tag_Name, '<2>', self._post_menu)
        self.cn.tag_bind(self.view_canvas.Clickable_Tag_Name, "<Double-1>", self._double_clicked)

    # =================================================================================================================
    # draw an individual block
    # =================================================================================================================
    def draw_block(self, resource_type: ResourceType, day: int, start_time: float,
                   duration: float, text: str, gui_block_id, movable=True):
        """
        Draws the specific block on the gui canvas
        :param resource_type:
        :param day:
        :param start_time:
        :param duration:
        :param text:  the text to display on the block
        :param gui_block_id:  the unique identifier for all gui objects that comprise this block
        :param movable: is this block allowed to be moved?
        """
        # draw the block
        self.view_canvas.draw_block(
            resource_type=resource_type,
            day = day, start_time = start_time, duration=duration,
            text=text,
            gui_tag=gui_block_id,
            movable=movable)

    # =================================================================================================================
    # modify movability of gui block
    # =================================================================================================================
    def modify_movable(self, gui_tag: str, movable: bool):
        """
        modify movability of gui block
        :param gui_tag:
        :param movable:
        :return:
        """
        self.view_canvas.modify_movable(gui_tag, movable)

    # =================================================================================================================
    # colour an individual block
    # =================================================================================================================
    def colour_block(self, gui_block_id: str, resource_type: ResourceType, is_movable=True,
                     conflict: ConflictType = None):
        """
        Set the colour of the block, base on its resource, conflict, etc.
        :param is_movable:
        :param conflict:
        :param gui_block_id:
        :param resource_type:
        :return:
        """

        conflict = ConflictType.most_severe(conflict, resource_type)
        colour = RESOURCE_COLOURS[resource_type]

        if conflict != ConflictType.NONE:
            colour = ConflictType.colours().get(conflict, "pink")

        if not is_movable:
            colour = IMMOVABLE_COLOUR

        cn = self.cn

        colour = Colour.string(colour)
        text_colour = "black"
        if not Colour.is_light(colour):
            text_colour = "white"

        # if tag name doesn't exist, then don't do anything
        try:
            cn.itemconfigure(f"{self.view_canvas.Rectange_Tag_Name} && {gui_block_id}", fill=colour)
            cn.itemconfigure(f"{self.view_canvas.Text_Tag_Name} && {gui_block_id}", fill=text_colour)
        except tk.TclError:
            pass

    # =================================================================================================================
    # put this gui window on top of every one else
    # =================================================================================================================
    def raise_to_top(self):
        """put this gui window on top of every one else"""
        self.toplevel.lift()

    # =================================================================================================================
    # the window is being closed (event handler for the little 'x' in top-level window)
    # =================================================================================================================
    def destroy(self):
        """Close/destroy the gui window."""
        self.on_closing_handler(self)
        self.toplevel.destroy()

    def kill(self):
        self.toplevel.destroy()
        self.mw.update_idletasks()

    # =================================================================================================================
    # menus, popups and double-clicks
    # =================================================================================================================
    def _post_menu(self, e: tk.Event):
        """
        Display pop up menu for specific gui block
        :param e: the Tk event that triggered this handler
        """
        gui_block_id = self.view_canvas.get_gui_block_id_from_selected_item()

        # get menu info from callback routine
        menu_details = self.get_popup_menu_handler(gui_block_id)
        if menu_details is None:
            return

        # create the menu and display
        menu = tk.Menu(self.toplevel, tearoff=0)
        generate_menu(self.toplevel, menu_details, menu)
        try:
            menu.tk_popup(e.x_root, e.y_root)
        finally:
            menu.grab_release()

    def _double_clicked(self, _: tk.Event):
        """
        call event handler because gui block was double-clicked
        """
        gui_block_id = self.view_canvas.get_gui_block_id_from_selected_item()
        self.double_click_block_handler(gui_block_id)


    # =================================================================================================================
    # undo/ redo
    # =================================================================================================================
    def _undo(self, _: tk.Event = None):
        """calls the "undo" handler"""
        self.undo_handler()

    def _redo(self, _: tk.Event = None):
        """calls the "redo" handler"""
        self.redo_handler()

    # ============================================================================
    # Moving Gui blocks around in response to changes in the schedule block
    # ============================================================================
    def move_gui_block(self, gui_block_id: str, day_number, start_number):
        """
        Move the gui block to a new location based on day
        :param gui_block_id: this tag specifies the gui tag associated with the drawn images
        :param day_number:
        :param start_number:
        """

        new_coords = self.view_canvas.get_coords(day_number, start_number)
        old_coords = self.view_canvas.gui_block_coords(gui_block_id)
        if old_coords is None:
            return

        (cur_x_pos, cur_y_pos, _, _) = old_coords
        self.cn.lift(gui_block_id)
        self.cn.move(gui_block_id, new_coords[0] - cur_x_pos, new_coords[1] - cur_y_pos)


    # ============================================================================
    # Dragging Guiblocks around
    # ============================================================================
    def _select_gui_block_to_move(self, event: tk.Event):

        gui_block_id = self.view_canvas.get_gui_block_id_from_selected_item()
        self.cn.tag_raise(gui_block_id, 'all')

        # unbind any previous binding for clicking and motion, just in case
        self.cn.bind("<Motion>", "")
        self.cn.bind("<ButtonRelease-1>", "")
        self.cn.tag_bind(self.view_canvas.Movable_Tag_Name, "<Button-1>", "")

        # bind for mouse motion
        self.cn.bind("<Motion>", partial(self._gui_block_is_moving, gui_block_id, event.x, event.y))

        # bind for release of mouse up
        self.cn.bind("<ButtonRelease-1>", partial(self._gui_block_has_stopped_moving, gui_block_id))

    def _gui_block_is_moving(self, gui_block_id, original_x, original_y, event: tk.Event):

        # unbind moving while we process
        self.cn.bind("<Motion>", "")

        # move the widget
        self.cn.move(gui_block_id, event.x - original_x, event.y - original_y)

        # get the information about the movement and call event handler
        info = self.view_canvas.gui_block_to_day_time_duration(gui_block_id)
        if info is not None:
            day, start_time, _ = info
            self.gui_block_is_moving_handler(gui_block_id, day, start_time)

        # rebind for motion
        self.cn.bind("<Motion>", partial(self._gui_block_is_moving, gui_block_id, event.x, event.y))

    def _gui_block_has_stopped_moving(self, gui_block_id, _: tk.Event):
        self.cn.tag_bind(self.view_canvas.Movable_Tag_Name, "<Button-1>", self._select_gui_block_to_move)
        self.cn.bind("<Motion>", "")
        self.cn.bind("<ButtonRelease-1>", "")
        info = self.view_canvas.gui_block_to_day_time_duration(gui_block_id)
        if info is not None:
            self.gui_block_has_dropped_handler(gui_block_id)




from .edit_resources_tk import EditResourcesTk
from .edit_resources_tk import DEColumnDescription
from .edit_courses_tk import EditCoursesTk
from .main_pages_tk import MainPageBaseTk
# from .scheduler_tk import SchedulerTk, set_main_page_event_handler, MAIN_PAGE_EVENT_HANDLERS



"""
Define a class time for teaching
"""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Optional
from ..Utilities.id_generator import IdGenerator
from .enums import WeekDay, ConflictType
from .time_slot import TimeSlot
OptionalId = Optional[int]

# stuff that we need just for type checking, not for actual functionality
if TYPE_CHECKING:
    from .lab import Lab
    from .teacher import Teacher
    from .section import Section
    from .stream import Stream

DEFAULT_DAY = WeekDay.Monday
DEFAULT_START = 8.0
DEFAULT_DURATION = 1.5

# =====================================================================================================================
# Block - Class time
# =====================================================================================================================
class Block:
    """
    Describes a block which is a specific time slot for teaching part of a section of a course.
    """

    # Class Variables
    block_ids = IdGenerator()

    # -----------------------------------------------------------------------------------------------------------------
    # Constructor
    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, section: Section,
                 day: WeekDay | float = DEFAULT_DAY,
                 start: float = DEFAULT_START,
                 duration: float = DEFAULT_DURATION,
                 movable: bool = True,
                 block_id: OptionalId = None) -> None:
        """
        Creates a new block object (class time)
        :param section: a block must be part of a course/section
        :param day:
        :param start: when does the class start (float)
        :param duration: how long does the class last
        :param movable: can the time/day of this class be moved
        :param block_id:
        """
        self._sync: list[Block] = []
        self.section = section
        self._time_slot = TimeSlot(day, start, duration, movable)

        self._teachers: set[Teacher] = set()
        self._labs: set[Lab] = set()
        self.conflict = ConflictType.NONE
        self._block_id = Block.block_ids.get_new_id(block_id)

    # -----------------------------------------------------------------------------------------------------------------
    # generic properites
    # -----------------------------------------------------------------------------------------------------------------
    @property
    def id(self):
        """block id"""
        return self._block_id

    @property
    def number(self):
        """block id"""
        return str(self._block_id)

    def description(self) -> str:
        """Returns text string that describes this Block."""
        return str(self._time_slot)

    # -----------------------------------------------------------------------------------------------------------------
    # time slot properties and functions
    # -----------------------------------------------------------------------------------------------------------------
    @property
    def start(self) -> float:
        return self._time_slot.start

    @start.setter
    def start(self, value: float):
        self._time_slot.start = value

    @property
    def day(self) -> WeekDay:
        return self._time_slot.day

    @day.setter
    def day(self, value: WeekDay|float):
        if isinstance(value,int) or isinstance(value,float):
            self._time_slot.day = WeekDay(int(value))
        else:
            self._time_slot.day = value

    @property
    def end(self):
        return self._time_slot.end

    @property
    def duration(self)->float:
        return self._time_slot.duration

    @duration.setter
    def duration(self, value: float):
        self._time_slot.duration = value

    @property
    def movable(self)->bool:
        return self._time_slot.movable

    @movable.setter
    def movable(self, value: bool):
        self._time_slot.movable = value

    def snap_to_time(self):
        self._time_slot.snap_to_time()

    def snap_to_day(self, fractional_day: float) -> bool:
        return self._time_slot.snap_to_day(fractional_day)

    def conflicts_time(self, other: Block) -> bool:
        """
        Tests if the current Block conflicts with another TimeSlot.
        :param other: other Block
        """
        return self._time_slot.conflicts_time(other._time_slot)

    # -----------------------------------------------------------------------------------------------------------------
    # Conflicts
    # -----------------------------------------------------------------------------------------------------------------
    def add_conflict(self, conflict: ConflictType):
        """add a conflict to any pre-existing conflict"""
        self.conflict |= conflict

    def clear_conflicts(self):
        """remove any and all conflicts"""
        self.conflict = ConflictType.NONE

    # -----------------------------------------------------------------------------------------------------------------
    # All about streams
    # -----------------------------------------------------------------------------------------------------------------
    def streams(self) -> tuple[Stream, ...]:
        """Returns an immutable list of the assigned to the section that this block is part of."""
        return self.section.streams()

    # -----------------------------------------------------------------------------------------------------------------
    # All about labs
    # -----------------------------------------------------------------------------------------------------------------
    def labs(self) -> tuple[Lab, ...]:
        """Returns an immutable list of the labs assigned to this block."""
        return tuple(sorted(self._labs))

    def add_lab(self, lab: Lab):
        """Assign a new lab, to this block"""
        self._labs.add(lab)

    def remove_lab(self, lab: Lab):
        """Removes the specified Lab from this Block."""
        self._labs.discard(lab)

    def remove_all_labs(self):
        """Removes ALL Labs from this Block."""
        self._labs.clear()

    def has_lab(self, lab: Lab) -> bool:
        """Returns true if the Block has the specified Lab."""
        return lab in self._labs

    # -----------------------------------------------------------------------------------------------------------------
    # All about Teachers
    # -----------------------------------------------------------------------------------------------------------------
    def teachers(self) -> tuple[Teacher, ...]:
        """Returns an immutable list of the teachers assigned to this block."""
        return tuple(sorted(self._teachers))

    def add_teacher(self, teacher: Teacher):
        """Assign a new teacher, to this block"""
        self._teachers.add(teacher)

    def remove_teacher(self, teacher: Teacher):
        """Removes the specified teacher from this Block."""
        self._teachers.discard(teacher)

    def remove_all_teachers(self):
        """Removes ALL Teachers from this Block."""
        self._teachers.clear()

    def has_teacher(self, teacher: Teacher) -> bool:
        """Returns true if the Block has the specified Lab."""
        return teacher in self._teachers

    # -----------------------------------------------------------------------------------------------------------------
    # All about syncing
    # -----------------------------------------------------------------------------------------------------------------
    def synced_blocks(self) -> tuple[Block, ...]:
        """Returns a tuple of the Blocks which are synced_blocks to this Block."""
        return tuple(self._sync)

    def sync_block(self, block: Block):
        """The new Block object will be synced_blocks with this one
        (i.e., changing the time_start time of this Block will change the time_start time of the
        synced_blocks block)."""
        self._sync.append(block)
        block._time_slot = self._time_slot
        for b in self._sync:
            b._sync.append(self)

    def unsync_block(self, block: Block):
        """Removes syncing of Block from this Block."""

        for b in self._sync:
            if self in b._sync:
                b._sync.remove(self)
        if block in self._sync:
            self._sync.remove(block)
            block._time_slot = copy.copy(block._time_slot)

    # -----------------------------------------------------------------------------------------------------------------
    # string representation of object
    # -----------------------------------------------------------------------------------------------------------------
    def __str__(self) -> str:
        """Returns a text string that describes the Block."""
        text = ""
        text += f"{self._time_slot} "
        text += ", ".join(str(lab.number) for lab in self.labs())
        text += f" for {self.section.title}"

        return text

    def __repr__(self) -> str:
        return f"{self.id} - {self.description()}"

    # -----------------------------------------------------------------------------------------------------------------
    # sorting properties
    # -----------------------------------------------------------------------------------------------------------------
    def __lt__(self,other):
        return self._time_slot < other._time_slot

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self._block_id)


